from ...backend_obj import backend, ArrayLike
from ... import config


def _h_poly(t: ArrayLike) -> ArrayLike:
    """Helper function to compute the 'h' polynomial matrix used in the
    cubic spline.

    Args:
        t (Tensor): A 1D tensor representing the normalized x values.

    Returns:
        Tensor: A 2D tensor of size (4, len(t)) representing the 'h' polynomial matrix.

    """

    tt = t[None, :] ** (backend.arange(4, device=config.DEVICE)[:, None])
    A = backend.as_array(
        [
            [1.0, 0.0, -3.0, 2.0],
            [0.0, 1.0, -2.0, 1.0],
            [0.0, 0.0, 3.0, -2.0],
            [0.0, 0.0, -1.0, 1.0],
        ],
        dtype=config.DTYPE,
        device=config.DEVICE,
    )
    return A @ tt


def cubic_spline_torch(x: ArrayLike, y: ArrayLike, xs: ArrayLike) -> ArrayLike:
    """Compute the 1D cubic spline interpolation for the given data points
    using PyTorch.

    **Args:**
    -  `x` (Tensor): A 1D tensor representing the x-coordinates of the known data points.
    -  `y` (Tensor): A 1D tensor representing the y-coordinates of the known data points.
    -  `xs` (Tensor): A 1D tensor representing the x-coordinates of the positions where
                the cubic spline function should be evaluated.
    """
    m = (y[1:] - y[:-1]) / (x[1:] - x[:-1])
    m = backend.concatenate([m[0].flatten(), (m[1:] + m[:-1]) / 2, m[-1].flatten()])
    idxs = backend.searchsorted(x[:-1], xs) - 1
    dx = x[idxs + 1] - x[idxs]
    hh = _h_poly((xs - x[idxs]) / dx)
    ret = hh[0] * y[idxs] + hh[1] * m[idxs] * dx + hh[2] * y[idxs + 1] + hh[3] * m[idxs + 1] * dx
    return ret


def spline(R: ArrayLike, profR: ArrayLike, profI: ArrayLike, extend: str = "zeros") -> ArrayLike:
    """Spline 1d profile function, cubic spline between points up
    to second last point beyond which is linear

    **Args:**
    -  `R`: Radii tensor at which to evaluate the spline function
    -  `profR`: radius values for the surface density profile in the same units as `R`
    -  `profI`: surface density values for the surface density profile
    -  `extend`: How to extend the spline beyond the last point. Options are 'zeros' or 'const'.
    """
    I = cubic_spline_torch(profR, profI, R.flatten()).reshape(*R.shape)
    if extend == "zeros":
        backend.fill_at_indices(I, R > profR[-1], 0)
    elif extend == "const":
        backend.fill_at_indices(I, R > profR[-1], profI[-1])
    else:
        raise ValueError(f"Unknown extend option: {extend}. Use 'zeros' or 'const'.")
    return I



import numpy as np
import warnings
from astropy.utils.data import download_file
from astropy.io import fits
from astropy.utils.exceptions import AstropyWarning
from numpy.core.defchararray import startswith

try:
    from pyvo.dal import sia
except:
    sia = None
import os

# Suppress common Astropy warnings that can clutter CI logs
warnings.simplefilter("ignore", category=AstropyWarning)


def flip_hdu(hdu):
    """
    Flips the image data in the FITS HDU on the RA axis to match the expected orientation.

    Args:
        hdu (astropy.io.fits.HDUList): The FITS HDU to be flipped.

    Returns:
        astropy.io.fits.HDUList: The flipped FITS HDU.
    """
    assert "CD1_1" in hdu[0].header, "HDU does not contain WCS information."
    assert "CD2_1" in hdu[0].header, "HDU does not contain WCS information."
    assert "CRPIX1" in hdu[0].header, "HDU does not contain WCS information."
    assert "NAXIS1" in hdu[0].header, "HDU does not contain WCS information."
    hdu[0].data = hdu[0].data[:, ::-1].copy()
    hdu[0].header["CD1_1"] = -hdu[0].header["CD1_1"]
    hdu[0].header["CD2_1"] = -hdu[0].header["CD2_1"]
    hdu[0].header["CRPIX1"] = int(hdu[0].header["NAXIS1"] / 2) + 1
    hdu[0].header["CRPIX2"] = int(hdu[0].header["NAXIS2"] / 2) + 1
    return hdu


def ls_open(ra, dec, size_arcsec, band="r", release="ls_dr9"):
    """
    Retrieves and opens a FITS cutout from the deepest stacked image in the
    specified Legacy Survey data release using the Astro Data Lab SIA service.

    Args:
        ra (float): Right Ascension in decimal degrees.
        dec (float): Declination in decimal degrees.
        size_arcsec (float): Size of the square cutout (side length) in arcseconds.
        band (str): The filter band (e.g., 'g', 'r', 'z'). Case-insensitive.
        release (str): The Legacy Survey Data Release (e.g., 'DR9').

    Returns:
        astropy.io.fits.HDUList: The opened FITS file object.
    """

    if sia is None:
        raise ImportError(
            "Cannot use ls_open without pyvo. Please install pyvo (pip install pyvo) before continuing."
        )

    # 1. Set the specific SIA service endpoint for the desired release
    # SIA endpoints for specific surveys are listed in the notebook.
    service_url = f"https://datalab.noirlab.edu/sia/{release.lower()}"
    svc = sia.SIAService(service_url)

    # 2. Convert size from arcseconds to degrees (FOV) for the SIA query
    # and apply the cosine correction for RA.
    fov_deg = size_arcsec / 3600.0

    # The search method takes the position (RA, Dec) and the square FOV.
    imgTable = svc.search(
        (ra, dec), (fov_deg / np.cos(dec * np.pi / 180.0), fov_deg), verbosity=2
    ).to_table()

    # 3. Filter the table for stacked images in the specified band
    target_band = band.lower()

    sel = (
        (imgTable["proctype"] == "Stack")
        & (imgTable["prodtype"] == "image")
        & (startswith(imgTable["obs_bandpass"].astype(str), target_band))
    )

    Table = imgTable[sel]

    if len(Table) == 0:
        raise ValueError(
            f"No stacked FITS image found for {release} band '{band}' at the requested RA {ra} and Dec {dec}."
        )

    # 4. Pick the "deepest" image (longest exposure time)
    # Note: 'exptime' data needs explicit float conversion for np.argmax
    max_exptime_index = np.argmax(Table["exptime"].data.data.astype("float"))
    row = Table[max_exptime_index]

    # 5. Download the file and open it
    url = row["access_url"]  # get the download URL

    # Use astropy's download_file, which handles the large data transfer
    # and automatically uses a long timeout (120s in the notebook example)
    filename = download_file(url, cache=False, show_progress=False, timeout=120)

    # Open the downloaded FITS file
    hdu = fits.open(filename)

    try:
        hdu = flip_hdu(hdu)
    except AssertionError:
        pass  # If WCS info is missing, skip flipping

    # Clean up the temporary file created by download_file
    try:
        os.remove(filename)
    except OSError:
        pass  # Ignore if cleanup fails

    return hdu



from .base import AstroPhotError

__all__ = ("InvalidWindow", "InvalidData", "InvalidImage")


class InvalidWindow(AstroPhotError):
    """
    Raised whenever a window is misspecified
    """


class InvalidData(AstroPhotError):
    """
    Raised when the data provided to an image is invalid or cannot be processed.
    """


class InvalidImage(AstroPhotError):
    """
    Raised when an image object cannot be used as given.
    """



from . import (
    conversions,
    initialize,
    decorators,
    integration,
    interpolate,
    parametric_profiles,
)
from .fitsopen import ls_open

__all__ = [
    "decorators",
    "interpolate",
    "integration",
    "parametric_profiles",
    "initialize",
    "conversions",
    "ls_open",
]



# Apply an optimizer toa  downsampled version of an image
from typing import Dict, Any

import numpy as np

from .base import BaseOptimizer
from ..models import Model
from .lm import LM
from .. import config

__all__ = ["MiniFit"]


class MiniFit(BaseOptimizer):
    """MiniFit optimizer that applies a fitting method to a downsampled version
    of the model's target image.

    This is useful for quickly optimizing parameters on a smaller scale before
    applying them to the full resolution image. With fewer pixels, the optimization
    can be faster and more efficient, especially for large images.

    This Optimizer can wrap any optimizer that follows the BaseOptimizer interface.

    **Args:**
    -  `downsample_factor`: Factor by which to downsample the target image. Default is 2.
    -  `max_pixels`: Maximum number of pixels in the downsampled image. Default is 10000.
    -  `method`: The optimizer method to use, e.g., `LM` for Levenberg-Marquardt. Default is `LM`.
    -  `method_kwargs`: Additional keyword arguments to pass to the optimizer method.
    """

    def __init__(
        self,
        model: Model,
        downsample_factor: int = 2,
        max_pixels: int = 10000,
        method: BaseOptimizer = LM,
        initial_state: np.ndarray = None,
        method_kwargs: Dict[str, Any] = {},
        **kwargs: Dict[str, Any],
    ) -> None:
        super().__init__(model, initial_state, **kwargs)

        self.method = method
        self.method_kwargs = method_kwargs
        if "verbose" not in self.method_kwargs:
            self.method_kwargs["verbose"] = self.verbose

        self.downsample_factor = downsample_factor
        self.max_pixels = max_pixels

    def fit(self) -> BaseOptimizer:
        initial_target = self.model.target
        target_area = self.model.target[self.model.window]
        while True:
            small_target = target_area.reduce(self.downsample_factor)
            if np.prod(small_target._data.shape) < self.max_pixels:
                break
            self.downsample_factor += 1

        if self.verbose > 0:
            config.logger.info(f"Downsampling target by {self.downsample_factor}x")

        self.small_target = small_target
        self.model.target = small_target
        res = self.method(self.model, **self.method_kwargs).fit()
        self.model.target = initial_target

        self.message = res.message

        return self



from functools import partial
from typing import Literal

import numpy as np
import torch
from scipy.stats import binned_statistic, iqr

from .. import config
from ..backend_obj import backend
from ..models import Model

from ..utils.conversions.units import flux_to_sb
from .visuals import *

__all__ = [
    "radial_light_profile",
    "radial_median_profile",
    "ray_light_profile",
    "wedge_light_profile",
    "warp_phase_profile",
]


def radial_light_profile(
    fig,
    ax,
    model: Model,
    rad_unit="arcsec",
    extend_profile=1.0,
    R0=0.0,
    resolution=1000,
    plot_kwargs={},
):
    """
    Used to plot the brightness profile as a function of radius for models which define a `radial_model`.

    **Args:**
    - `fig`: matplotlib figure object
    - `ax`: matplotlib axis object
    - `model` (Model): Model object from which to plot the radial profile.
    - `rad_unit` (str): The name of the radius units to plot. If you select "pixel" then the plot will work in pixel units (physical radii divided by pixelscale) if you choose any other string then it will remain in the physical units of the image and the axis label will be whatever you set the value to. Default: "arcsec". Options: "arcsec", "pixel"
    - `extend_profile` (float): The factor by which to extend the profile beyond the maximum radius of the model's window. Default: 1.0
    - `R0` (float): The starting radius for the profile. Default: 0.0
    - `resolution` (int): The number of points to use in the profile. Default: 1000
    - `plot_kwargs` (dict): Additional keyword arguments to pass to the plot function, such as `linewidth`, `color`, etc.
    """
    xx = backend.linspace(
        R0,
        max(model.window.shape) * backend.to_numpy(model.target.pixelscale) * extend_profile / 2,
        int(resolution),
        dtype=config.DTYPE,
        device=config.DEVICE,
    )
    flux = backend.to_numpy(model.radial_model(xx, params=()))
    if model.target.zeropoint is not None:
        yy = flux_to_sb(flux, 1.0, model.target.zeropoint.item())
    else:
        yy = np.log10(flux)

    kwargs = {
        "linewidth": 2,
        "color": main_pallet["primary1"],
        "label": f"{model.name} profile",
    }
    kwargs.update(plot_kwargs)
    ax.plot(
        backend.to_numpy(xx),
        yy,
        **kwargs,
    )

    if model.target.zeropoint is not None:
        ax.set_ylabel("Surface Brightness [mag/arcsec$^2$]")
        if not ax.yaxis_inverted():
            ax.invert_yaxis()
    else:
        ax.set_ylabel("log$_{10}$(flux/arcsec$^2$)")
    ax.set_xlabel(f"Radius [{rad_unit}]")
    ax.set_xlim([R0, None])
    return fig, ax


def radial_median_profile(
    fig,
    ax,
    model: Model,
    count_limit: int = 10,
    return_profile: bool = False,
    rad_unit: str = "arcsec",
    plot_kwargs: dict = {},
):
    """
    Plot an SB profile by taking flux median at each radius.

    Using the coordinate transforms defined by the model object,
    assigns a radius to each pixel then bins the pixel-radii and
    computes the median in each bin. This gives a simple
    representation of the image data if one were to simply average the
    pixels along isophotes.

    **Args:**
    -  `fig`: matplotlib figure object
    -  `ax`: matplotlib axis object
    -  `model` (AstroPhot_Model): Model object from which to determine the radial binning. Also provides the target image to extract the data
    -  `count_limit` (int): The limit of pixels in a bin, below which uncertainties are not computed. Default: 10
    -  `return_profile` (bool): Instead of just returning the fig and ax object, will return the extracted profile formatted as: Rbins (the radial bin edges), medians (the median in each bin), scatter (the 16-84 quartile range / 2), count (the number of pixels in each bin). Default: False
    -  `rad_unit` (str): The name of the radius units to plot. If you select "pixel" then the plot will work in pixel units (physical radii divided by pixelscale) if you choose any other string then it will remain in the physical units of the image and the axis label will be whatever you set the value to. Default: "arcsec". Options: "arcsec", "pixel"
    -  `plot_kwargs` (dict): Additional keyword arguments to pass to the plot function, such as `linewidth`, `color`, etc.

    """

    Rlast_pix = max(model.window.shape) / 2
    Rlast_phys = Rlast_pix * model.target.pixelscale.item()

    Rbins = [0.0]
    while Rbins[-1] < Rlast_phys:
        Rbins.append(Rbins[-1] + max(2 * model.target.pixelscale.item(), Rbins[-1] * 0.1))
    Rbins = np.array(Rbins)

    with torch.no_grad():
        image = model.target[model.window]
        x, y = image.coordinate_center_meshgrid()
        x, y = model.transform_coordinates(x, y, params=())
        R = backend.sqrt(x**2 + y**2)
        R = backend.to_numpy(R)

    dat = backend.to_numpy(image._data)
    # remove masked pixels
    mask = backend.to_numpy(image._mask)
    dat = dat[~mask]
    R = R[~mask]

    count, bins, binnum = binned_statistic(
        R.ravel(),
        dat.ravel(),
        statistic="count",
        bins=Rbins,
    )

    stat, bins, binnum = binned_statistic(
        R.ravel(),
        dat.ravel(),
        statistic="median",
        bins=Rbins,
    )
    stat[count == 0] = np.nan

    scat, bins, binnum = binned_statistic(
        R.ravel(),
        dat.ravel(),
        statistic=partial(iqr, rng=(16, 84)),
        bins=Rbins,
    )
    scat[count > count_limit] /= 2 * np.sqrt(count[count > count_limit])
    scat[count <= count_limit] = 0

    if model.target.zeropoint is not None:
        stat = flux_to_sb(stat, model.target.pixel_area.item(), model.target.zeropoint.item())
        ax.set_ylabel("Surface Brightness [mag/arcsec$^2$]")
        if not ax.yaxis_inverted():
            ax.invert_yaxis()
    else:
        stat = np.log10(stat)
        ax.set_ylabel("log$_{10}$(flux/arcsec^2)")

    kwargs = {
        "linewidth": 0,
        "elinewidth": 1,
        "color": main_pallet["primary2"],
        "label": "data profile",
        **plot_kwargs,
    }
    ax.errorbar(
        (Rbins[:-1] + Rbins[1:]) / 2,
        stat,
        yerr=scat,
        fmt=".",
        **kwargs,
    )
    ax.set_xlabel(f"Radius [{rad_unit}]")

    if return_profile:
        return Rbins, stat, scat, count
    return fig, ax


def ray_light_profile(
    fig,
    ax,
    model: Model,
    rad_unit="arcsec",
    extend_profile=1.0,
    resolution=1000,
):
    """
    Used for plotting ray (wedge) type models which define a `iradial_model` method. These have multiple radial profiles.

    **Args:**
    - `fig`: matplotlib figure object
    - `ax`: matplotlib axis object
    - `model` (Model): Model object from which to plot the radial profile.
    - `rad_unit` (str): The name of the radius units to plot.
    - `extend_profile` (float): The factor by which to extend the profile beyond the maximum radius of the model's window. Default: 1.0
    - `resolution` (int): The number of points to use in the profile. Default: 1000
    """
    xx = backend.linspace(
        0,
        max(model.window.shape) * model.target.pixelscale * extend_profile / 2,
        int(resolution),
        dtype=config.DTYPE,
        device=config.DEVICE,
    )
    for r in range(model.segments):
        if model.segments <= 3:
            col = main_pallet[f"primary{r+1}"]
        else:
            col = cmap_grad(r / model.segments)
        with torch.no_grad():
            ax.plot(
                backend.to_numpy(xx),
                np.log10(backend.to_numpy(model.iradial_model(r, xx, params=()))),
                linewidth=2,
                color=col,
                label=f"{model.name} profile {r}",
            )
    ax.set_ylabel("log$_{10}$(flux)")
    ax.set_xlabel(f"Radius [{rad_unit}]")

    return fig, ax


def warp_phase_profile(fig, ax, model: Model, rad_unit="arcsec"):
    """Used to plot the phase profile of a warp model. This gives the axis ratio and position angle as a function of radius."""
    ax.plot(
        backend.to_numpy(model.q_R.prof),
        model.q_R.npvalue,
        linewidth=2,
        color=main_pallet["primary1"],
        label=f"{model.name} axis ratio",
    )
    ax.plot(
        backend.to_numpy(model.PA_R.prof),
        model.PA_R.npvalue / np.pi,
        linewidth=2,
        color=main_pallet["primary2"],
        label=f"{model.name} position angle/$\\pi$",
    )
    ax.set_ylim([0, 1])
    ax.set_ylabel("q [b/a], PA [rad/$\\pi$]")
    ax.set_xlabel(f"Radius [{rad_unit}]")

    return fig, ax



from typing import Optional, Literal

import numpy as np

from ...param import forward
from ...backend_obj import backend, ArrayLike
from ... import config
from ...image import Image, Window, JacobianImage
from .. import func
from ...errors import SpecificationConflict


class SampleMixin:
    """
    **Options:**
    -    `sampling_mode`: The method used to sample the model in image pixels. Options are:
            - `auto`: Automatically choose the sampling method based on the image size.
            - `midpoint`: Use midpoint sampling, evaluate the brightness at the center of each pixel.
            - `simpsons`: Use Simpson's rule for sampling integrating each pixel.
            - `quad:x`: Use quadrature sampling with order x, where x is a positive integer to integrate each pixel.
    -    `jacobian_maxparams`: The maximum number of parameters before the Jacobian will be broken into smaller chunks. This is helpful for limiting the memory requirements to build a model.
    -    `jacobian_maxpixels`: The maximum number of pixels before the Jacobian will be broken into smaller chunks. This is helpful for limiting the memory requirements to build a model.
    -    `integrate_mode`: The method used to select pixels to integrate further where the model varies significantly. Options are:
            - `none`: No extra integration is performed (beyond the sampling_mode).
            - `bright`: Select the brightest pixels for further integration.
            - `threshold`: Select pixels which show signs of significant higher order derivatives.
    -    `integrate_tolerance`: The tolerance for selecting a pixel in the integration method. This is the total flux fraction that is integrated over the image.
    -    `integrate_fraction`: The fraction of the pixels to super sample during integration.
    -    `integrate_max_depth`: The maximum depth of the integration method.
    -    `integrate_gridding`: The gridding used for the integration method to super-sample a pixel at each iteration.
    -    `integrate_quad_order`: The order of the quadrature used for the integration method on the super sampled pixels.
    """

    # Method for initial sampling of model
    sampling_mode = "auto"  # auto (choose based on image size), midpoint, simpsons, quad:x (where x is a positive integer)

    # Maximum size of parameter list before jacobian will be broken into smaller chunks, this is helpful for limiting the memory requirements to build a model, lower jacobian_chunksize is slower but uses less memory
    jacobian_maxparams = 10
    jacobian_maxpixels = 1000**2
    integrate_mode = "bright"  # none, bright, curvature
    integrate_fraction = 0.05  # fraction of the pixels to super sample
    integrate_max_depth = 2
    integrate_gridding = 5
    integrate_quad_order = 3

    _options = (
        "sampling_mode",
        "jacobian_maxparams",
        "jacobian_maxpixels",
        "integrate_mode",
        "integrate_fraction",
        "integrate_max_depth",
        "integrate_gridding",
        "integrate_quad_order",
    )

    @forward
    def _bright_integrate(self, sample: ArrayLike, image: Image) -> ArrayLike:
        i, j = image.pixel_center_meshgrid()
        sample = func.bright_integrate(
            sample,
            i,
            j,
            lambda i, j: self.brightness(*image.pixel_to_plane(i, j)),
            scale=image.base_scale,
            bright_frac=self.integrate_fraction,
            quad_order=self.integrate_quad_order,
            gridding=self.integrate_gridding,
            max_depth=self.integrate_max_depth,
        )
        return sample

    @forward
    def _curvature_integrate(self, sample: ArrayLike, image: Image) -> ArrayLike:
        i, j = image.pixel_center_meshgrid()
        kernel = func.curvature_kernel(config.DTYPE, config.DEVICE)
        curvature = (
            backend.abs(
                backend.pad(
                    backend.conv2d(
                        sample.reshape(1, 1, *sample.shape),
                        kernel.reshape(1, 1, *kernel.shape),
                        padding="valid",
                    ),
                    (0, 0, 0, 0, 1, 1, 1, 1),
                    mode="replicate",
                )
            )
            .squeeze(0)
            .squeeze(0)
        )
        N = max(1, int(np.prod(image.data.shape) * self.integrate_fraction))
        select = backend.topk(curvature.flatten(), N)[1]

        sample_flat = sample.flatten()
        sample_flat = backend.fill_at_indices(
            sample_flat,
            select,
            func.recursive_quad_integrate(
                i.flatten()[select],
                j.flatten()[select],
                lambda i, j: self.brightness(*image.pixel_to_plane(i, j)),
                scale=image.base_scale,
                curve_frac=self.integrate_fraction,
                quad_order=self.integrate_quad_order,
                gridding=self.integrate_gridding,
                max_depth=self.integrate_max_depth,
            ),
        )
        return sample_flat.reshape(sample.shape)

    @forward
    def sample_image(self, image: Image) -> ArrayLike:
        if self.sampling_mode == "auto":
            N = np.prod(image._data.shape[:2])
            if N <= 100:
                sampling_mode = "quad:5"
            elif N <= 10000:
                sampling_mode = "simpsons"
            else:
                sampling_mode = "midpoint"
        else:
            sampling_mode = self.sampling_mode
        if sampling_mode == "midpoint":
            x, y = image.coordinate_center_meshgrid()
            res = self.brightness(x, y)
            sample = func.pixel_center_integrator(res)
        elif sampling_mode == "simpsons":
            x, y = image.coordinate_simpsons_meshgrid()
            res = self.brightness(x, y)
            sample = func.pixel_simpsons_integrator(res)
        elif sampling_mode.startswith("quad:"):
            order = int(self.sampling_mode.split(":")[1])
            i, j, w = image.pixel_quad_meshgrid(order=order)
            x, y = image.pixel_to_plane(i, j)
            res = self.brightness(x, y)
            sample = func.pixel_quad_integrator(res, w)
        else:
            raise SpecificationConflict(
                f"Unknown sampling mode {self.sampling_mode} for model {self.name}"
            )
        if self.integrate_mode == "curvature":
            sample = self._curvature_integrate(sample, image)
        elif self.integrate_mode == "bright":
            sample = self._bright_integrate(sample, image)
        elif self.integrate_mode != "none":
            raise SpecificationConflict(
                f"Unknown integrate mode {self.integrate_mode} for model {self.name}"
            )
        return sample

    def _jacobian(
        self, window: Window, params_pre: ArrayLike, params: ArrayLike, params_post: ArrayLike
    ) -> ArrayLike:
        # return jacfwd( # this should be more efficient, but the trace overhead is too high
        #     lambda x: self.sample(
        #         window=window, params=torch.cat((params_pre, x, params_post), dim=-1)
        #     ).data
        # )(params)
        return backend.jacobian(
            lambda x: self.sample(
                window=window, params=backend.concatenate((params_pre, x, params_post), dim=-1)
            )._data,
            params,
        )

    def jacobian(
        self,
        window: Optional[Window] = None,
        pass_jacobian: Optional[JacobianImage] = None,
        params: Optional[ArrayLike] = None,
    ) -> JacobianImage:
        if window is None:
            window = self.window

        if pass_jacobian is None:
            jac_img = self.target[window].jacobian_image(
                parameters=self.build_params_array_identities()
            )
        else:
            jac_img = pass_jacobian

        # No dynamic params
        if params is None:
            params = self.get_values()
        if params.shape[-1] == 0:
            return jac_img

        # handle large images
        n_pixels = np.prod(window.shape)
        if n_pixels > self.jacobian_maxpixels:
            for chunk in window.chunk(self.jacobian_maxpixels):
                jac_img = self.jacobian(window=chunk, pass_jacobian=jac_img, params=params)
            return jac_img

        identities = self.build_params_array_identities()
        if len(jac_img.match_parameters(identities)[0]) == 0:
            return jac_img

        target = self.target[window]
        if len(params) > self.jacobian_maxparams:  # handle large number of parameters
            chunksize = len(params) // self.jacobian_maxparams + 1
            for i in range(0, len(params), chunksize):
                params_pre = params[:i]
                params_chunk = params[i : i + chunksize]
                params_post = params[i + chunksize :]
                jac_chunk = self._jacobian(window, params_pre, params_chunk, params_post)
                jac_img += target.jacobian_image(
                    parameters=identities[i : i + chunksize],
                    data=jac_chunk,
                )
        else:
            jac = self._jacobian(window, params[:0], params, params[0:0])
            jac_img += target.jacobian_image(parameters=identities, data=jac)

        return jac_img

    def gradient(
        self,
        window: Optional[Window] = None,
        params: Optional[ArrayLike] = None,
        likelihood: Literal["gaussian", "poisson"] = "gaussian",
    ) -> ArrayLike:
        """Compute the gradient of the model with respect to its parameters."""
        if window is None:
            window = self.window

        jacobian_image = self.jacobian(window=window, params=params)

        data = self.target[window]._data
        model = self.sample(window=window)._data
        if likelihood == "gaussian":
            weight = self.target[window]._weight
            gradient = backend.sum(
                jacobian_image._data * ((data - model) * weight)[..., None], dim=(0, 1)
            )
        elif likelihood == "poisson":
            gradient = backend.sum(
                jacobian_image._data * (1 - data / model)[..., None],
                dim=(0, 1),
            )

        return gradient



from .lm import LM, LMfast
from .gradient import Grad, Slalom
from .iterative import Iter, IterParam
from .scipy_fit import ScipyFit
from .minifit import MiniFit
from .hmc import HMC
from .mala import MALA
from .mhmcmc import MHMCMC
from . import func

__all__ = [
    "LM",
    "LMfast",
    "Grad",
    "Iter",
    "MALA",
    "IterParam",
    "ScipyFit",
    "MiniFit",
    "HMC",
    "MHMCMC",
    "Slalom",
    "func",
]



from typing import List, Optional

import numpy as np

from .image_object import Image
from .jacobian_image import JacobianImage
from .. import config
from ..backend_obj import backend, ArrayLike
from .mixins import DataMixin

__all__ = ["PSFImage"]


class PSFImage(DataMixin, Image):
    """Image object which represents a model of PSF (Point Spread Function).

    PSFImage inherits from the base Image class and represents the model of a point spread function.
    The point spread function characterizes the response of an imaging system to a point source or point object.

    The shape of the PSF data should be odd (for your sanity) but this is not enforced.
    """

    def __init__(self, *args, **kwargs):
        kwargs.update({"crval": (0, 0), "crpix": (0, 0), "crtan": (0, 0)})
        super().__init__(*args, **kwargs)
        self.crpix = (np.array(self._data.shape[:2], dtype=np.float64) - 1.0) / 2

    def normalize(self):
        """Normalizes the PSF image to have a sum of 1."""
        norm = backend.sum(self._data)
        self._data = self._data / norm
        self._weight = self._weight * norm**2

    @property
    def psf_pad(self) -> int:
        return max(self._data.shape[:2]) // 2

    def jacobian_image(
        self,
        parameters: Optional[List[str]] = None,
        data: Optional[ArrayLike] = None,
        **kwargs,
    ) -> JacobianImage:
        """
        Construct a blank `JacobianImage` object formatted like this current `PSFImage` object. Mostly used internally.
        """
        if parameters is None:
            data = None
            parameters = []
        elif data is None:
            data = backend.zeros(
                (*self._data.shape, len(parameters)),
                dtype=config.DTYPE,
                device=config.DEVICE,
            )
        kwargs = {
            "CD": self.CD.value,
            "crpix": self.crpix,
            "crtan": self.crtan.value,
            "crval": self.crval.value,
            "zeropoint": self.zeropoint,
            "identity": self.identity,
            **kwargs,
        }
        return JacobianImage(parameters=parameters, _data=data, **kwargs)

    def model_image(self, **kwargs) -> "PSFImage":
        """
        Construct a blank `ModelImage` object formatted like this current `TargetImage` object. Mostly used internally.
        """
        kwargs = {
            "_data": backend.zeros_like(self._data),
            "CD": self.CD.value,
            "crpix": self.crpix,
            "crtan": self.crtan.value,
            "crval": self.crval.value,
            "identity": self.identity,
            **kwargs,
        }
        return PSFImage(**kwargs)

    @property
    def zeropoint(self):
        return None

    @zeropoint.setter
    def zeropoint(self, value):
        """PSFImage does not support zeropoint."""
        pass

    def plane_to_world(self, x, y):
        raise NotImplementedError(
            "PSFImage does not support plane_to_world conversion. There is no meaningful world position of a PSF image."
        )

    def world_to_plane(self, ra, dec):
        raise NotImplementedError(
            "PSFImage does not support world_to_plane conversion. There is no meaningful world position of a PSF image."
        )



import numpy as np
from numpy import ndarray
from .conversions.functions import sersic_n_to_b

__all__ = (
    "sersic_np",
    "gaussian_np",
    "exponential_np",
    "moffat_np",
    "nuker_np",
    "ferrer_np",
    "king_np",
)


def sersic_np(R: ndarray, n: ndarray, Re: ndarray, Ie: ndarray) -> ndarray:
    """Sersic 1d profile function, works more generally with numpy
    operations. In the event that impossible values are passed to the
    function it returns large values to guide optimizers away from
    such values.

    **Args:**
    -  `R`: Radii array at which to evaluate the sersic function
    -  `n`: sersic index restricted to n > 0.36
    -  `Re`: Effective radius in the same units as R
    -  `Ie`: Effective surface density
    """
    if np.any(np.array([n, Re, Ie]) <= 0):
        return np.ones(len(R)) * 1e6
    bn = sersic_n_to_b(n)
    return Ie * np.exp(-bn * ((R / Re) ** (1 / n) - 1))


def gaussian_np(R: ndarray, sigma: ndarray, I0: ndarray) -> ndarray:
    """Gaussian 1d profile function, works more generally with numpy
    operations.

    **Args:**
    -  `R`: Radii array at which to evaluate the gaussian function
    -  `sigma`: standard deviation of the gaussian in the same units as R
    -  `I0`: central surface density
    """
    return (I0 / np.sqrt(2 * np.pi * sigma**2)) * np.exp(-0.5 * ((R / sigma) ** 2))


def exponential_np(R: ndarray, Ie: ndarray, Re: ndarray) -> ndarray:
    """Exponential 1d profile function, works more generally with numpy
    operations.

    **Args:**
    -  `R`: Radii array at which to evaluate the exponential function
    -  `Ie`: Effective surface density
    -  `Re`: Effective radius in the same units as R
    """
    return Ie * np.exp(-sersic_n_to_b(1.0) * (R / Re - 1.0))


def moffat_np(R: ndarray, n: ndarray, Rd: ndarray, I0: ndarray) -> ndarray:
    """Moffat 1d profile function, works with numpy operations.

    **Args:**
    -  `R`: Radii array at which to evaluate the moffat function
    -  `n`: concentration index
    -  `Rd`: scale length in the same units as R
    -  `I0`: central surface density
    """
    return I0 / (1 + (R / Rd) ** 2) ** n


def nuker_np(
    R: ndarray, Rb: ndarray, Ib: ndarray, alpha: ndarray, beta: ndarray, gamma: ndarray
) -> ndarray:
    """Nuker 1d profile function, works with numpy functions

    **Args:**
    -  `R`: Radii tensor at which to evaluate the nuker function
    -  `Ib`: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
    -  `Rb`: scale length radius
    -  `alpha`: sharpness of transition between power law slopes
    -  `beta`: outer power law slope
    -  `gamma`: inner power law slope

    """
    return (
        Ib
        * (2 ** ((beta - gamma) / alpha))
        * ((R / Rb) ** (-gamma))
        * ((1 + (R / Rb) ** alpha) ** ((gamma - beta) / alpha))
    )


def ferrer_np(R: ndarray, rout: ndarray, alpha: ndarray, beta: ndarray, I0: ndarray) -> ndarray:
    """
    Modified Ferrer profile.

    **Args:**
    -  `R`: Radial distance from the center.
    -  `rout`: Outer radius of the profile.
    -  `alpha`: Power-law index.
    -  `beta`: Exponent for the modified Ferrer function.
    -  `I0`: Central intensity.
    """
    return (R < rout) * I0 * ((1 - (np.clip(R, 0, rout) / rout) ** (2 - beta)) ** alpha)


def king_np(R: ndarray, Rc: ndarray, Rt: ndarray, alpha: ndarray, I0: ndarray) -> ndarray:
    """
    Empirical King profile.

    **Args:**
    -  `R`: The radial distance from the center.
    -  `Rc`: The core radius of the profile.
    -  `Rt`: The truncation radius of the profile.
    -  `alpha`: The power-law index of the profile.
    -  `I0`: The central intensity of the profile.
    """
    beta = 1 / (1 + (Rt / Rc) ** 2) ** (1 / alpha)
    gamma = 1 / (1 + (R / Rc) ** 2) ** (1 / alpha)
    return (R < Rt) * I0 * ((np.clip(gamma, 0, 1) - beta) / (1 - beta)) ** alpha



import numpy as np

from ...errors import OptimizeStopFail, OptimizeStopSuccess
from ...backend_obj import backend
from ... import config


def nll(D, M, W):
    """
    Negative log-likelihood for Gaussian noise.
    D: data
    M: model prediction
    W: weights
    """
    return 0.5 * backend.sum(W * (D - M) ** 2)


def nll_poisson(D, M):
    """
    Negative log-likelihood for Poisson noise.
    D: data
    M: model prediction
    """
    return backend.sum(M - D * backend.log(M + 1e-10))  # Adding small value to avoid log(0)


def gradient(J, W, D, M):
    return J.T @ (W * (D - M))[:, None]


def gradient_poisson(J, D, M):
    return J.T @ (D / M - 1)[:, None]


def hessian(J, W):
    return J.T @ (W[:, None] * J)


def hessian_poisson(J, D, M):
    return J.T @ ((D / (M**2 + 1e-10))[:, None] * J)


def damp_hessian(hess, L):
    I = backend.eye(len(hess), dtype=config.DTYPE, device=config.DEVICE)
    D = backend.ones_like(hess) - I
    return hess * (I + D / (1 + L)) + L * I * backend.diag(hess)


def solve(hess, grad, L):
    hessD = damp_hessian(hess, L)  # (N, N)
    while True:
        try:
            h = backend.linalg.solve(hessD, grad)
            break
        except backend.LinAlgErr:
            hessD = hessD + L * backend.eye(len(hessD), dtype=config.DTYPE, device=config.DEVICE)
            L = L * 2
    return hessD, h


def lm_step(
    x,
    data,
    model,
    weight,
    jacobian,
    L=1.0,
    Lup=9.0,
    Ldn=11.0,
    tolerance=1e-4,
    likelihood="gaussian",
):
    L0 = L
    M0 = backend.detach(model(x))  # (M,)
    J = backend.detach(jacobian(x))  # (M, N)

    if likelihood == "gaussian":
        nll0 = nll(data, M0, weight).item()
        grad = gradient(J, weight, data, M0)  # (N, 1)
        hess = hessian(J, weight)  # (N, N)
    elif likelihood == "poisson":
        nll0 = nll_poisson(data, M0).item()
        grad = gradient_poisson(J, data, M0)  # (N, 1)
        hess = hessian_poisson(J, data, M0)  # (N, N)
    else:
        raise ValueError(f"Unsupported likelihood: {likelihood}")

    del J

    if backend.allclose(grad, backend.zeros_like(grad)):
        raise OptimizeStopSuccess("Gradient is zero, optimization converged.")

    best = {"x": backend.zeros_like(x), "nll": nll0, "L": L}
    scary = {"x": None, "nll": np.inf, "L": None, "rho": np.inf}
    nostep = True
    improving = None
    for i in range(10):
        hessD, h = solve(hess, grad, L)  # (N, N), (N, 1)
        M1 = model(x + h.squeeze(1))  # (M,)
        if likelihood == "gaussian":
            nll1 = nll(data, M1, weight).item()
        elif likelihood == "poisson":
            nll1 = nll_poisson(data, M1).item()

        # Handle nan chi2
        if not np.isfinite(nll1):
            L *= Lup
            if improving is True:
                break
            improving = False
            continue

        if backend.allclose(h, backend.zeros_like(h)) and L < 0.1:
            if i == 0:
                raise OptimizeStopSuccess("Step with zero length means optimization complete.")
            break

        # actual nll improvement vs expected from linearization
        rho = (nll0 - nll1) / backend.abs(h.T @ hessD @ h - 2 * grad.T @ h).item()

        if (nll1 < (nll0 + tolerance) and abs(rho - 1) < abs(scary["rho"] - 1)) or (
            nll1 < scary["nll"] and rho > -10
        ):
            scary = {"x": x + h.squeeze(1), "nll": nll1, "L": L0, "rho": rho}

        # Avoid highly non-linear regions
        if rho < 0.1 or rho > 2:
            L *= Lup
            if improving is True:
                break
            improving = False
            continue

        if nll1 < best["nll"]:  # new best
            best = {"x": x + h.squeeze(1), "nll": nll1, "L": L}
            nostep = False
            L /= Ldn
            if L < 1e-8 or improving is False:
                break
            improving = True
        elif improving is True:  # were improving, now not improving
            break
        else:  # not improving and bad chi2, damp more
            L *= Lup
            if L >= 1e9:
                break
            improving = False

        # If we are improving chi2 by more than 10% then we can stop
        if (best["nll"] - nll0) / nll0 < -0.1:
            break

    if nostep:
        if scary["x"] is not None and (scary["nll"] - nll0) / nll0 < tolerance:
            return scary
        raise OptimizeStopFail("Could not find step to improve chi^2")

    return best



from ...backend_obj import ArrayLike


def moffat(R: ArrayLike, n: ArrayLike, Rd: ArrayLike, I0: ArrayLike) -> ArrayLike:
    """Moffat 1d profile function

    **Args:**
    -  `R`: Radii tensor at which to evaluate the moffat function
    -  `n`: concentration index
    -  `Rd`: scale length in the same units as R
    -  `I0`: central surface density

    """
    return I0 / (1 + (R / Rd) ** 2) ** n



__all__ = ("AstroPhotError", "SpecificationConflict")


class AstroPhotError(Exception):
    """
    Base exception for all AstroPhot processes.
    """


class SpecificationConflict(AstroPhotError):
    """
    Raised when the inputs to an object are conflicting and/or ambiguous
    """



import numpy as np

from ...errors import OptimizeStopFail, OptimizeStopSuccess
from ...backend_obj import backend


def slalom_step(f, g, x0, m, S, N=10, up=1.3, down=0.5):
    l = [f(x0).item()]
    d = [0.0]
    grad = g(x0)
    if backend.allclose(grad, backend.zeros_like(grad)):
        raise OptimizeStopSuccess("success: Gradient is zero, optimization converged.")

    D = grad + m
    D = D / backend.linalg.norm(D)
    seeking = False
    for _ in range(N):
        l.append(f(x0 - S * D).item())
        d.append(S)

        # Check if the last value is finite
        if not np.isfinite(l[-1]):
            l.pop()
            d.pop()
            S *= down
            continue

        if seeking and np.argmin(l) == len(l) - 1:
            # If we are seeking a minimum and the last value is the minimum, we can stop
            break

        if len(l) < 3:
            # Seek better step size based on loss improvement
            if l[-1] < l[-2]:
                S *= up
            else:
                S *= down
        else:
            O = np.polyfit(d[-3:], l[-3:], 2)
            if O[0] > 0:
                S = -O[1] / (2 * O[0])
                seeking = True
            else:
                S *= down
                seeking = False

    if np.argmin(l) == 0:
        raise OptimizeStopFail("fail: cannot find step to improve.")
    return d[np.argmin(l)], l[np.argmin(l)], grad



from ...backend_obj import ArrayLike


def nuker(
    R: ArrayLike,
    Rb: ArrayLike,
    Ib: ArrayLike,
    alpha: ArrayLike,
    beta: ArrayLike,
    gamma: ArrayLike,
) -> ArrayLike:
    """Nuker 1d profile function

    **Args:**
    -  `R`: Radii tensor at which to evaluate the nuker function
    -  `Ib`: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
    -  `Rb`: scale length radius
    -  `alpha`: sharpness of transition between power law slopes
    -  `beta`: outer power law slope
    -  `gamma`: inner power law slope

    """
    return (
        Ib
        * (2 ** ((beta - gamma) / alpha))
        * ((R / Rb) ** (-gamma))
        * ((1 + (R / Rb) ** alpha) ** ((gamma - beta) / alpha))
    )



from . import config, models, plots, utils, fit, image, errors
from .param import forward, Param, Module

from .image import (
    Image,
    ImageList,
    TargetImage,
    TargetImageList,
    SIPModelImage,
    SIPTargetImage,
    CMOSModelImage,
    CMOSTargetImage,
    JacobianImage,
    JacobianImageList,
    PSFImage,
    ModelImage,
    ModelImageList,
    Window,
    WindowList,
)
from .models import Model
from .backend_obj import backend, ArrayLike

try:
    from ._version import version as VERSION  # noqa
except ModuleNotFoundError:
    VERSION = "0.0.0"
    print(
        "WARNING: AstroPhot version number not found. This is likely because you are running AstroPhot from a source directory."
    )


# meta data
__version__ = VERSION
__author__ = "Connor Stone"
__email__ = "connorstone628@gmail.com"

__all__ = (
    "models",
    "image",
    "Model",
    "Image",
    "ImageList",
    "TargetImage",
    "TargetImageList",
    "SIPModelImage",
    "SIPTargetImage",
    "CMOSModelImage",
    "CMOSTargetImage",
    "JacobianImage",
    "JacobianImageList",
    "PSFImage",
    "ModelImage",
    "ModelImageList",
    "Window",
    "WindowList",
    "plots",
    "utils",
    "fit",
    "forward",
    "Param",
    "errors",
    "Module",
    "config",
    "backend",
    "ArrayLike",
    "__version__",
    "__author__",
    "__email__",
)



from typing import Tuple
from ...backend_obj import backend, ArrayLike


def rotate(theta: ArrayLike, x: ArrayLike, y: ArrayLike) -> Tuple[ArrayLike, ArrayLike]:
    """
    Applies a rotation matrix to the X,Y coordinates
    """
    s = backend.sin(theta)
    c = backend.cos(theta)
    return c * x - s * y, s * x + c * y



from .target_image import TargetImage
from .mixins import CMOSMixin
from .model_image import ModelImage
from ..backend_obj import backend
from .. import config


class CMOSModelImage(CMOSMixin, ModelImage):
    """A ModelImage with CMOS-specific functionality."""

    def fluxdensity_to_flux(self):
        # CMOS pixels only sensitive in sub area, so scale the flux density
        self._data = self._data * self.pixel_area * self.subpixel_scale**2


class CMOSTargetImage(CMOSMixin, TargetImage):
    """
    A TargetImage with CMOS-specific functionality.
    This class is used to represent a target image with CMOS-specific features.
    It inherits from TargetImage and CMOSMixin.
    """

    def model_image(self, upsample: int = 1, pad: int = 0, **kwargs) -> CMOSModelImage:
        """Model the image with CMOS-specific features."""
        if upsample > 1 or pad > 0:
            raise NotImplementedError("Upsampling and padding are not implemented for CMOS images.")

        kwargs = {
            "subpixel_loc": self.subpixel_loc,
            "subpixel_scale": self.subpixel_scale,
            "_data": backend.zeros(self._data.shape[:2], dtype=config.DTYPE, device=config.DEVICE),
            "CD": self.CD.value,
            "crpix": self.crpix,
            "crtan": self.crtan.value,
            "crval": self.crval.value,
            "zeropoint": self.zeropoint,
            "identity": self.identity,
            "name": self.name + "_model",
            **kwargs,
        }
        return CMOSModelImage(**kwargs)



from ...utils.integration import quad_table
from ...backend_obj import backend, ArrayLike


def pixel_center_meshgrid(shape: tuple[int, int], dtype, device) -> tuple:
    i = backend.arange(shape[0], dtype=dtype, device=device)
    j = backend.arange(shape[1], dtype=dtype, device=device)
    return backend.meshgrid(i, j, indexing="ij")


def cmos_pixel_center_meshgrid(
    shape: tuple[int, int], loc: tuple[float, float], dtype, device
) -> tuple:
    i = backend.arange(shape[0], dtype=dtype, device=device) + loc[0]
    j = backend.arange(shape[1], dtype=dtype, device=device) + loc[1]
    return backend.meshgrid(i, j, indexing="ij")


def pixel_corner_meshgrid(shape: tuple[int, int], dtype, device) -> tuple:
    i = backend.arange(shape[0] + 1, dtype=dtype, device=device) - 0.5
    j = backend.arange(shape[1] + 1, dtype=dtype, device=device) - 0.5
    return backend.meshgrid(i, j, indexing="ij")


def pixel_simpsons_meshgrid(shape: tuple[int, int], dtype, device) -> tuple:
    i = 0.5 * backend.arange(2 * shape[0] + 1, dtype=dtype, device=device) - 0.5
    j = 0.5 * backend.arange(2 * shape[1] + 1, dtype=dtype, device=device) - 0.5
    return backend.meshgrid(i, j, indexing="ij")


def pixel_quad_meshgrid(shape: tuple[int, int], dtype, device, order=3) -> tuple:
    i, j = pixel_center_meshgrid(shape, dtype, device)
    di, dj, w = quad_table(order, dtype, device)
    i = backend.repeat(i[..., None], order**2, -1) + di.flatten()
    j = backend.repeat(j[..., None], order**2, -1) + dj.flatten()
    return i, j, w.flatten()


def rotate(theta: ArrayLike, x: ArrayLike, y: ArrayLike) -> tuple:
    """
    Applies a rotation matrix to the X,Y coordinates
    """
    s = theta.sin()
    c = theta.cos()
    return c * x - s * y, s * x + c * y



from .base import AstroPhotError, SpecificationConflict
from .fit import OptimizeStopFail, OptimizeStopSuccess
from .image import InvalidWindow, InvalidData, InvalidImage
from .models import InvalidTarget, UnrecognizedModel

__all__ = (
    "AstroPhotError",
    "SpecificationConflict",
    "OptimizeStopFail",
    "OptimizeStopSuccess",
    "InvalidWindow",
    "InvalidData",
    "InvalidImage",
    "InvalidTarget",
    "UnrecognizedModel",
)



# Apply a different optimizer iteratively
from typing import Dict, Any, Union, Sequence, Literal
from time import time
from functools import partial

from caskade import ValidContext
import numpy as np
import torch

from .base import BaseOptimizer
from ..models import Model
from .lm import LM
from .. import config
from ..backend_obj import backend
from ..errors import OptimizeStopSuccess, OptimizeStopFail
from . import func

__all__ = [
    "Iter",
    # "Iter_LM"
]


class Iter(BaseOptimizer):
    """Optimizer wrapper that performs optimization iteratively.

    This optimizer applies the LM optimizer to a group model iteratively one
    model at a time. It can be used for complex fits or when the number of
    models to fit is too large to fit in memory. Note that it will iterate
    through the group model, but if models within the group are themselves group
    models, then they will be optimized as a whole. This gives some flexibility
    to structure the models in a useful way.

    If not given, the `lm_kwargs` will be set to a relative tolerance of 1e-3
    and a maximum of 15 iterations. This is to allow for faster convergence, it
    is not worthwhile for a single model to spend lots of time optimizing when
    its neighbors havent converged.

    **Args:**
    -    `max_iter`: Maximum number of iterations, defaults to 100.
    -    `lm_kwargs`: Keyword arguments to pass to `LM` optimizer.
    """

    def __init__(
        self,
        model: Model,
        initial_state: np.ndarray = None,
        max_iter: int = 100,
        lm_kwargs: Dict[str, Any] = {"verbose": 0},
        **kwargs: Dict[str, Any],
    ):
        super().__init__(model, initial_state, max_iter=max_iter, **kwargs)

        self.current_state = model.get_values()
        self.lm_kwargs = lm_kwargs
        if "relative_tolerance" not in lm_kwargs:
            # Lower tolerance since it's not worth fine tuning a model when its neighbors will be shifting soon anyway
            self.lm_kwargs["relative_tolerance"] = 1e-3
            self.lm_kwargs["max_iter"] = 15
        #          # pixels      # parameters
        self.ndf = self.model.target[self.model.window].flatten("data").shape[0] - len(
            self.current_state
        )
        # subtract masked pixels from degrees of freedom
        self.ndf -= backend.sum(self.model.target[self.model.window].flatten("mask")).item()

    def sub_step(self, model: Model, update_uncertainty=False):
        """
        Perform optimization for a single model.
        """
        self.Y -= model()
        initial_values = model.target.copy()
        model.target = model.target - self.Y
        res = LM(model, **self.lm_kwargs).fit(update_uncertainty=update_uncertainty)
        self.Y += model()
        if self.verbose > 1:
            config.logger.info(res.message)
        model.target = initial_values

    def step(self):
        """
        Perform a single iteration of optimization.
        """
        if self.verbose > 0:
            config.logger.info("--------iter-------")

        # Fit each model individually
        for model in self.model.models:
            if self.verbose > 0:
                config.logger.info(model.name)
            self.sub_step(model)
        # Update the current state
        self.current_state = self.model.get_values()

        # Update the loss value
        with torch.no_grad():
            if self.verbose > 0:
                config.logger.info("Update Chi^2 with new parameters")
            self.Y = self.model(params=self.current_state)
            D = self.model.target[self.model.window].flatten("data")
            V = self.model.target[self.model.window].flatten("variance")
            M = self.model.target[self.model.window].flatten("mask")
            loss = backend.sum((((D - self.Y.flatten("data")) ** 2) / V)[~M]) / self.ndf
        if self.verbose > 0:
            config.logger.info(f"Loss: {loss.item()}")
        self.lambda_history.append(np.copy(backend.to_numpy(self.current_state)))
        self.loss_history.append(loss.item())

        # Test for convergence
        if self.iteration >= 2 and (
            (-self.relative_tolerance * 1e-3)
            < ((self.loss_history[-2] - self.loss_history[-1]) / self.loss_history[-1])
            < (self.relative_tolerance / 10)
        ):
            self._count_finish += 1
        else:
            self._count_finish = 0

        self.iteration += 1

    def fit(self) -> BaseOptimizer:
        """
        Perform the iterative fitting process until convergence or maximum iterations reached.
        """
        self.iteration = 0
        self.Y = self.model(params=self.current_state)
        start_fit = time()
        try:
            while True:
                self.step()
                if self.iteration > 2 and self._count_finish >= 2:
                    self.message = self.message + "success"
                    break
                elif self.iteration >= self.max_iter:
                    self.message = self.message + f"fail max iterations reached: {self.iteration}"
                    break

        except KeyboardInterrupt:
            self.message = self.message + "fail interrupted"

        self.model.set_values(
            backend.as_array(self.res(), dtype=config.DTYPE, device=config.DEVICE)
        )
        if self.verbose > 1:
            config.logger.info(
                f"Iter Fitting complete in {time() - start_fit} sec with message: {self.message}"
            )

        return self


class IterParam(BaseOptimizer):
    """Optimization wrapper that call LM optimizer on subsets of variables.

    IterParam takes the full set of parameters for a model and breaks them down
    into chunks as specified by the user. It then calls Levenberg-Marquardt
    optimization on the subset of parameters, and iterates through all subsets
    until every parameter has been optimized. It cycles through these chunks
    until convergence. This method is very powerful in situations where the full
    optimization problem cannot fit in memory, or where the optimization problem
    is too complex to tackle as a single large problem. In full LM optimization
    a single problematic parameter can ripple into issues with every other
    parameter, so breaking the problem down can sometimes make an otherwise
    intractable problem easier. For small problems with only a few models, it is
    likely better to optimize the full problem with LM as, when it works, LM is
    faster than the IterParam method.

    Args:
      chunks (Union[int, tuple]): Specify how to break down the model
        parameters. If an integer, at each iteration the algorithm will break the
        parameters into groups of that size. If a tuple, should be a tuple of
        arrays of length num_dimensions which act as selectors for the parameters
        to fit (1 to include, 0 to exclude). Default: 50
      chunk_order (str): How to iterate through the chunks. Should be one of: random,
        sequential. Default: sequential
    """

    def __init__(
        self,
        model: Model,
        initial_state: Sequence = None,
        chunks: Union[int, tuple] = 50,
        chunk_order: Literal["random", "sequential"] = "sequential",
        max_iter: int = 100,
        relative_tolerance: float = 1e-5,
        Lup=11.0,
        Ldn=9.0,
        L0=1.0,
        max_step_iter: int = 10,
        ndf=None,
        W=None,
        likelihood="gaussian",
        **kwargs,
    ):

        super().__init__(
            model,
            initial_state,
            max_iter=max_iter,
            relative_tolerance=relative_tolerance,
            **kwargs,
        )
        # Maximum number of iterations of the algorithm
        self.max_iter = max_iter
        # Maximum number of steps while searching for chi^2 improvement on a single jacobian evaluation
        self.max_step_iter = max_step_iter
        self.Lup = Lup
        self.Ldn = Ldn
        self.L = L0
        self.likelihood = likelihood
        if self.likelihood not in ["gaussian", "poisson"]:
            raise ValueError(f"Unsupported likelihood: {self.likelihood}")
        self.chunks = self.make_chunks(chunks)
        self.chunk_order = chunk_order

        # mask
        fit_mask = self.model.fit_mask()
        if isinstance(fit_mask, tuple):
            fit_mask = backend.concatenate(tuple(FM.flatten() for FM in fit_mask))
        else:
            fit_mask = fit_mask.flatten()

        mask = self.model.target[self.fit_window].flatten("mask")
        mask = mask | fit_mask
        self.mask = ~mask

        if backend.sum(self.mask).item() == 0:
            raise OptimizeStopSuccess("No data to fit. All pixels are masked")

        # Initialize optimizer attributes
        self.Y = self.model.target[self.fit_window].flatten("data")[self.mask]

        # 1 / (sigma^2)
        if W is not None:
            self.W = backend.as_array(W, dtype=config.DTYPE, device=config.DEVICE).flatten()[
                self.mask
            ]
        else:
            self.W = self.model.target[self.fit_window].flatten("weight")[self.mask]

        # The forward model which computes the output image given input parameters
        self.full_forward = lambda x: model(window=self.fit_window, params=x).flatten("data")[
            self.mask
        ]
        self.forward = []
        # Compute the jacobian
        self.jacobian = []

        f = lambda c, state, x: model(
            window=self.fit_window,
            params=backend.fill_at_indices(backend.copy(state), self.chunks[c], x),
        ).flatten("data")[self.mask]
        j = backend.jacfwd(
            lambda c, state, x: self.model(
                window=self.fit_window,
                params=backend.fill_at_indices(backend.copy(state), self.chunks[c], x),
            ).flatten("data")[self.mask],
            argnums=2,
        )
        for c in range(len(self.chunks)):
            self.forward.append(partial(f, c))
            self.jacobian.append(partial(j, c))

        # variable to store covariance matrix if it is ever computed
        self._covariance_matrix = None

        # Degrees of freedom
        if ndf is None:
            self.ndf = max(1.0, len(self.Y) - len(self.current_state))
        else:
            self.ndf = ndf

    def make_chunks(self, chunks):
        if isinstance(chunks, int):
            new_chunks = []
            for i in range(0, len(self.current_state), chunks):
                chunk = np.zeros(len(self.current_state), dtype=bool)
                chunk[i : i + chunks] = True
                new_chunks.append(chunk)
            chunks = new_chunks
        return chunks

    def iter_chunks(self):
        if self.chunk_order == "random":
            chunk_ids = list(range(len(self.chunks)))
            np.random.shuffle(chunk_ids)
        elif self.chunk_order == "sequential":
            chunk_ids = list(range(len(self.chunks)))
        else:
            raise ValueError(
                f"Unrecognized chunk_order: {self.chunk_order}. Should be one of: random, sequential"
            )
        return chunk_ids

    def chi2_ndf(self):
        return (
            backend.sum(self.W * (self.Y - self.full_forward(self.current_state)) ** 2) / self.ndf
        )

    def poisson_2nll_ndf(self):
        M = self.full_forward(self.current_state)
        return 2 * backend.sum(M - self.Y * backend.log(M + 1e-10)) / self.ndf

    @torch.no_grad()
    def fit(self, update_uncertainty=True) -> BaseOptimizer:
        """This performs the fitting operation. It iterates the LM step
        function until convergence is reached. Includes a message
        after fitting to indicate how the fitting exited. Typically if
        the message returns a "success" then the algorithm found a
        minimum. This may be the desired solution, or a pathological
        local minimum, this often depends on the initial conditions.

        """

        if len(self.current_state) == 0:
            if self.verbose > 0:
                config.logger.warning("No parameters to optimize. Exiting fit")
            self.message = "No parameters to optimize. Exiting fit"
            return self

        if self.likelihood == "gaussian":
            quantity = "Chi^2/DoF"
            self.loss_history = [self.chi2_ndf().item()]
        elif self.likelihood == "poisson":
            quantity = "2NLL/DoF"
            self.loss_history = [self.poisson_2nll_ndf().item()]
        self._covariance_matrix = None
        self.L_history = [self.L]
        self.lambda_history = [backend.to_numpy(backend.copy(self.current_state))]
        if self.verbose > 0:
            config.logger.info(
                f"==Starting LM fit for '{self.model.name}' with {len(self.current_state)} dynamic parameters and {len(self.Y)} pixels=="
            )

        for _ in range(self.max_iter):
            # Report status
            if self.verbose > 0:
                config.logger.info(f"{quantity}: {self.loss_history[-1]:.6g}, L: {self.L:.3g}")

            # Perform fitting
            chunk_L = []
            for c in self.iter_chunks():
                try:
                    if self.fit_valid:
                        with ValidContext(self.model):
                            valid_state = self.model.to_valid(self.current_state)
                            res = func.lm_step(
                                x=valid_state[self.chunks[c]],
                                data=self.Y,
                                model=partial(self.forward[c], valid_state),
                                weight=self.W,
                                jacobian=partial(self.jacobian[c], valid_state),
                                L=self.L,
                                Lup=self.Lup,
                                Ldn=self.Ldn,
                                likelihood=self.likelihood,
                            )
                        self.current_state = self.model.from_valid(
                            backend.fill_at_indices(
                                valid_state, self.chunks[c], backend.copy(res["x"])
                            )
                        )
                    else:
                        res = func.lm_step(
                            x=self.current_state[self.chunks[c]],
                            data=self.Y,
                            model=partial(self.forward[c], self.current_state),
                            weight=self.W,
                            jacobian=partial(self.jacobian[c], self.current_state),
                            L=self.L,
                            Lup=self.Lup,
                            Ldn=self.Ldn,
                            likelihood=self.likelihood,
                        )
                        self.current_state = backend.fill_at_indices(
                            self.current_state, self.chunks[c], backend.copy(res["x"])
                        )
                except OptimizeStopFail:
                    if self.verbose > 0:
                        config.logger.warning(
                            f"Could not find step to improve Chi^2 on chunk {c}, moving to next chunk"
                        )
                    continue
                except OptimizeStopSuccess as e:
                    continue  # success on individual chunk is not enough to stop overall fit
                chunk_L.append(res["L"])

            # Record progress
            self.L = np.clip(np.max(chunk_L), 1e-9, 1e9)
            self.L_history.append(self.L)
            self.loss_history.append(2 * res["nll"] / self.ndf)
            self.lambda_history.append(backend.to_numpy(backend.copy(self.current_state)))
            if self.check_convergence():
                break

        else:
            self.message = self.message + "fail. Maximum iterations"

        if self.verbose > 0:
            config.logger.info(
                f"Final {quantity}: {np.nanmin(self.loss_history):.6g}, L: {self.L_history[np.nanargmin(self.loss_history)]:.3g}. Converged: {self.message}"
            )

        self.model.set_values(
            backend.as_array(self.res(), dtype=config.DTYPE, device=config.DEVICE)
        )
        if update_uncertainty:
            self.update_uncertainty()

        return self

    def check_convergence(self) -> bool:
        """Check if the optimization has converged based on the last
        iteration's chi^2 and the relative tolerance.
        """
        if len(self.loss_history) < 3:
            return False
        good_history = [self.loss_history[0]]
        for l in self.loss_history[1:]:
            if good_history[-1] > l:
                good_history.append(l)
        if len(self.loss_history) - len(good_history) >= 10:
            self.message = self.message + "success by immobility. Convergence not guaranteed"
            return True
        if len(good_history) < 3:
            return False
        if (good_history[-2] - good_history[-1]) / good_history[
            -1
        ] < self.relative_tolerance and self.L < 0.1:
            self.message = self.message + "success"
            return True
        if len(good_history) < 10:
            return False
        if (good_history[-10] - good_history[-1]) / good_history[-1] < self.relative_tolerance:
            self.message = self.message + "success by immobility. Convergence not guaranteed"
            return True
        return False

    @property
    @torch.no_grad()
    def covariance_matrix(self):
        """The covariance matrix for the model at the current
        parameters. This can be used to construct a full Gaussian PDF for the
        parameters using: $\\mathcal{N}(\\mu,\\Sigma)$ where $\\mu$ is the
        optimized parameters and $\\Sigma$ is the covariance matrix.

        """

        if self._covariance_matrix is not None:
            return self._covariance_matrix

        N = len(self.current_state)
        self._covariance_matrix = backend.zeros((N, N), dtype=config.DTYPE, device=config.DEVICE)
        for c in self.iter_chunks():
            J = self.jacobian[c](self.current_state, self.current_state[self.chunks[c]])
            if self.likelihood == "gaussian":
                hess = func.hessian(J, self.W)
            elif self.likelihood == "poisson":
                hess = func.hessian_poisson(J, self.Y, self.full_forward(self.current_state))
            try:
                sub_covariance_matrix = backend.linalg.inv(hess)
            except:
                config.logger.warning(
                    "WARNING: Hessian is singular, likely at least one parameter is non-physical. Will use pseudo-inverse of Hessian to continue but results should be inspected."
                )
                sub_covariance_matrix = backend.linalg.pinv(hess)

            ids = backend.meshgrid(
                backend.as_array(np.arange(N)[self.chunks[c]], dtype=int, device=config.DEVICE),
                backend.as_array(np.arange(N)[self.chunks[c]], dtype=int, device=config.DEVICE),
                indexing="ij",
            )
            self._covariance_matrix = backend.fill_at_indices(
                self._covariance_matrix, (ids[0], ids[1]), sub_covariance_matrix
            )
        return self._covariance_matrix

    @torch.no_grad()
    def update_uncertainty(self) -> None:
        """Call this function after optimization to set the uncertainties for
        the parameters. This will use the diagonal of the covariance
        matrix to update the uncertainties. See the covariance_matrix
        function for the full representation of the uncertainties.

        """
        # set the uncertainty for each parameter
        cov = self.covariance_matrix
        if backend.all(backend.isfinite(cov)):
            try:
                self.model.set_values(
                    backend.sqrt(backend.abs(backend.diag(cov))), attribute="uncertainty"
                )
            except RuntimeError as e:
                config.logger.warning(f"Unable to update uncertainty due to: {e}")
        else:
            config.logger.warning(
                "Unable to update uncertainty due to non finite covariance matrix"
            )



from ...backend_obj import backend, ArrayLike


def king(R: ArrayLike, Rc: ArrayLike, Rt: ArrayLike, alpha: ArrayLike, I0: ArrayLike) -> ArrayLike:
    """
    Empirical King profile.

    **Args:**
    -  `R`: Radial distance from the center of the profile.
    -  `Rc`: Core radius of the profile.
    -  `Rt`: Truncation radius of the profile.
    -  `alpha`: Power-law index of the profile.
    -  `I0`: Central intensity of the profile.
    """
    beta = 1 / (1 + (Rt / Rc) ** 2) ** (1 / alpha)
    gamma = 1 / (1 + (R / Rc) ** 2) ** (1 / alpha)
    return backend.where(
        R < Rt,
        I0 * ((backend.clamp(gamma, 0, 1) - beta) / (1 - beta)) ** alpha,
        backend.zeros_like(R),
    )



from typing import Sequence, Optional

import numpy as np
from scipy.optimize import minimize
from scipy.special import gammainc

from .. import config
from ..backend_obj import backend, ArrayLike
from ..models import Model
from ..image import Window


__all__ = ("BaseOptimizer",)


class BaseOptimizer:
    """
    Base optimizer object that other optimizers inherit from. Ensures consistent signature for the classes.

    **Args:**
    -  `model`: an AstroPhot_Model object that will have its (unlocked) parameters optimized [AstroPhot_Model]
    -  `initial_state`: optional initialization for the parameters as a 1D tensor [tensor]
    -  `relative_tolerance`: tolerance for counting success steps as: $0 < (\\chi_2^2 - \\chi_1^2)/\\chi_1^2 < \\text{tol}$ [float]
    -  `fit_window`: optional window to fit the model on [Window]
    -  `verbose`: verbosity level for the optimizer [int]
    -  `max_iter`: maximum allowed number of iterations [int]
    -  `save_steps`: optional string for path to save the model at each step (fitter dependent), e.g. "model_step_{step}.hdf5" [str]
    -  `fit_valid`: whether to fit while forcing parameters into valid range, or allow any value for each parameter. Default True [bool]

    """

    def __init__(
        self,
        model: Model,
        initial_state: Sequence = None,
        relative_tolerance: float = 1e-3,
        fit_window: Optional[Window] = None,
        verbose: int = 1,
        max_iter: int = None,
        save_steps: Optional[str] = None,
        fit_valid: bool = True,
    ) -> None:

        self.model = model
        self.verbose = verbose

        if initial_state is None:
            self.current_state = model.get_values()
        else:
            self.current_state = backend.as_array(
                initial_state, dtype=config.DTYPE, device=config.DEVICE
            )

        if fit_window is None:
            self.fit_window = self.model.window
        else:
            self.fit_window = fit_window & self.model.window

        self.max_iter = max_iter if max_iter is not None else 100 * len(self.current_state)
        self.iteration = 0
        self.save_steps = save_steps
        self.fit_valid = fit_valid

        self.relative_tolerance = relative_tolerance
        self.lambda_history = []
        self.loss_history = []
        self.message = ""

    def fit(self) -> "BaseOptimizer":
        raise NotImplementedError("Please use a subclass of BaseOptimizer for optimization")

    def step(self, current_state: ArrayLike = None) -> None:
        raise NotImplementedError("Please use a subclass of BaseOptimizer for optimization")

    def chi2min(self) -> float:
        """
        Returns the minimum value of chi^2 loss in the loss history.
        """
        return np.nanmin(self.loss_history)

    def res(self) -> np.ndarray:
        """Returns the value of lambda (state parameters) at which minimum loss was achieved."""
        N = np.isfinite(self.loss_history)
        if np.sum(N) == 0:
            config.logger.warning(
                "Getting optimizer res with no real loss history, using current state"
            )
            return backend.to_numpy(self.current_state)
        return np.array(self.lambda_history)[N][np.argmin(np.array(self.loss_history)[N])]

    def res_loss(self):
        """returns the minimum value from the loss history."""
        N = np.isfinite(self.loss_history)
        return np.min(np.array(self.loss_history)[N])



from typing import Optional
import numpy as np

__all__ = (
    "deg_to_arcsec",
    "arcsec_to_deg",
    "flux_to_sb",
    "flux_to_mag",
    "sb_to_flux",
    "mag_to_flux",
    "magperarcsec2_to_mag",
    "mag_to_magperarcsec2",
    "PA_shift_convention",
)

deg_to_arcsec = 3600.0
arcsec_to_deg = 1.0 / deg_to_arcsec


def flux_to_sb(flux: float, pixel_area: float, zeropoint: float) -> float:
    """Conversion from flux units to logarithmic surface brightness
    units.

    $$\\mu = -2.5\\log_{10}(flux) + z.p. + 2.5\\log_{10}(A)$$

    where $z.p.$ is the zeropoint and $A$ is the area of a pixel.

    """
    return -2.5 * np.log10(flux) + zeropoint + 2.5 * np.log10(pixel_area)


def flux_to_mag(flux: float, zeropoint: float, fluxe: Optional[float] = None) -> float:
    """Converts a flux total into logarithmic magnitude units.

    $$m = -2.5\\log_{10}(flux) + z.p.$$

    where $z.p.$ is the zeropoint.

    """
    if fluxe is None:
        return -2.5 * np.log10(flux) + zeropoint
    else:
        return -2.5 * np.log10(flux) + zeropoint, 2.5 * fluxe / (np.log(10) * flux)


def sb_to_flux(sb: float, pixel_area: float, zeropoint: float) -> float:
    """Converts logarithmic surface brightness units into flux units.

    $$flux = A 10^{-(\\mu - z.p.)/2.5}$$

    where $z.p.$ is the zeropoint and $A$ is the area of a pixel.

    """
    return pixel_area * 10 ** (-(sb - zeropoint) / 2.5)


def mag_to_flux(mag: float, zeropoint: float, mage: Optional[float] = None) -> float:
    """converts logarithmic magnitude units into a flux total.

    $$flux = 10^{-(m - z.p.)/2.5}$$

    where $z.p.$ is the zeropoint.

    """
    if mage is None:
        return 10 ** (-(mag - zeropoint) / 2.5)
    else:
        I = 10 ** (-(mag - zeropoint) / 2.5)
        return I, np.log(10) * I * mage / 2.5


def magperarcsec2_to_mag(
    mu: float, a: Optional[float] = None, b: Optional[float] = None, A: Optional[float] = None
) -> float:
    """
    Converts mag/arcsec^2 to mag

    **Args:**
    - `mu`: mag/arcsec^2
    - `a`: semi major axis radius (arcsec)
    - `b`: semi minor axis radius (arcsec)
    - `A`: pre-calculated area (arcsec^2)


    $$m = \\mu -2.5\\log_{10}(A)$$

    where $A$ is an area in arcsec^2.

    """
    assert (A is not None) or (a is not None and b is not None)
    if A is None:
        A = np.pi * a * b
    return mu - 2.5 * np.log10(
        A
    )  # https://en.wikipedia.org/wiki/Surface_brightness#Calculating_surface_brightness


def mag_to_magperarcsec2(
    m: float,
    a: Optional[float] = None,
    b: Optional[float] = None,
    R: Optional[float] = None,
    A: Optional[float] = None,
) -> float:
    """
    Converts mag to mag/arcsec^2

    **Args:**
    - `m`: mag
    - `a`: semi major axis radius (arcsec)
    - `b`: semi minor axis radius (arcsec)
    - `A`: pre-calculated area (arcsec^2)


    $$\\mu = m + 2.5\\log_{10}(A)$$

    where $A$ is an area in arcsec^2.
    """
    assert (A is not None) or (a is not None and b is not None) or (R is not None)
    if R is not None:
        A = np.pi * (R**2)
    elif A is None:
        A = np.pi * a * b
    return m + 2.5 * np.log10(
        A
    )  # https://en.wikipedia.org/wiki/Surface_brightness#Calculating_surface_brightness



from typing import Tuple
import numpy as np

from ...utils.integration import quad_table
from ...backend_obj import backend, ArrayLike
from ... import config


def pixel_center_integrator(Z: ArrayLike) -> ArrayLike:
    return Z


def pixel_simpsons_integrator(Z: ArrayLike) -> ArrayLike:
    kernel = (
        backend.as_array(
            [[[[1, 4, 1], [4, 16, 4], [1, 4, 1]]]], dtype=config.DTYPE, device=config.DEVICE
        )
        / 36.0
    )
    Z = backend.conv2d(Z.reshape(1, 1, *Z.shape), kernel, padding="valid", stride=2)
    return Z.squeeze(0).squeeze(0)


def pixel_quad_integrator(Z: ArrayLike, w: ArrayLike = None, order: int = 3) -> ArrayLike:
    """
    Integrate the pixel values using quadrature weights.

    **Args:**
    -  `Z`: The tensor containing pixel values.
    -  `w`: The quadrature weights.
    -  `order`: The order of the quadrature.
    """
    if w is None:
        _, _, w = quad_table(order, config.DTYPE, config.DEVICE)
    Z = Z * w
    return backend.sum(Z, dim=-1)


def upsample(i: ArrayLike, j: ArrayLike, order: int, scale: float) -> Tuple[ArrayLike, ArrayLike]:
    dp = (
        backend.linspace(-1, 1, order, dtype=config.DTYPE, device=config.DEVICE)
        * (order - 1)
        / (2.0 * order)
    )
    di, dj = backend.meshgrid(dp, dp, indexing="xy")

    si = backend.repeat(i[..., None], order**2, -1) + scale * di.flatten()
    sj = backend.repeat(j[..., None], order**2, -1) + scale * dj.flatten()
    return si, sj


def single_quad_integrate(
    i: ArrayLike, j: ArrayLike, brightness_ij, scale: float, quad_order: int = 3
) -> Tuple[ArrayLike, ArrayLike]:
    di, dj, w = quad_table(quad_order, config.DTYPE, config.DEVICE)
    qi = backend.repeat(i[..., None], quad_order**2, -1) + scale * di.flatten()
    qj = backend.repeat(j[..., None], quad_order**2, -1) + scale * dj.flatten()
    z = brightness_ij(qi, qj)
    z0 = backend.mean(z, dim=-1)
    z = backend.sum(z * w.flatten(), dim=-1)
    return z, z0


def recursive_quad_integrate(
    i: ArrayLike,
    j: ArrayLike,
    brightness_ij: callable,
    curve_frac: float,
    scale: float = 1.0,
    quad_order: int = 3,
    gridding: int = 5,
    _current_depth: int = 0,
    max_depth: int = 1,
) -> ArrayLike:
    z, z0 = single_quad_integrate(i, j, brightness_ij, scale, quad_order)

    if _current_depth >= max_depth:
        return z

    N = max(1, int(np.prod(z.shape) * curve_frac))
    select = backend.topk(backend.abs(z - z0).flatten(), N)[1]

    integral_flat = z.flatten()

    si, sj = upsample(i.flatten()[select], j.flatten()[select], gridding, scale)

    integral_flat = backend.fill_at_indices(
        integral_flat,
        select,
        backend.mean(
            recursive_quad_integrate(
                si,
                sj,
                brightness_ij,
                curve_frac=curve_frac,
                scale=scale / gridding,
                quad_order=quad_order,
                gridding=gridding,
                _current_depth=_current_depth + 1,
                max_depth=max_depth,
            ),
            dim=-1,
        ),
    )

    return integral_flat.reshape(z.shape)


def bright_integrate(
    z: ArrayLike,
    i: ArrayLike,
    j: ArrayLike,
    brightness_ij: callable,
    bright_frac: float,
    scale: float = 1.0,
    quad_order: int = 3,
    gridding: int = 5,
    max_depth: int = 2,
):
    trace = []
    for d in range(max_depth):
        N = max(1, int(np.prod(z.shape) * bright_frac))
        z_flat = z.flatten()
        select = backend.topk(z_flat, N)[1]
        trace.append([z_flat, select, z.shape])
        if d > 0:
            i, j = upsample(i.flatten()[select], j.flatten()[select], gridding, scale)
            scale = scale / gridding
        else:
            i, j = i.flatten()[select].reshape(-1, 1), j.flatten()[select].reshape(-1, 1)
        z, _ = single_quad_integrate(i, j, brightness_ij, scale, quad_order)
    trace.append([z, None, z.shape])

    for _ in reversed(range(1, max_depth + 1)):
        T = trace.pop(-1)
        trace[-1][0] = backend.fill_at_indices(
            trace[-1][0], trace[-1][1], backend.mean(T[0].reshape(T[2]), dim=-1)
        )

    return trace[0][0].reshape(trace[0][2])



from typing import Union
import numpy as np
from scipy.special import gamma
from ...backend_obj import backend, ArrayLike

__all__ = (
    "sersic_n_to_b",
    "sersic_I0_to_flux_np",
    "sersic_flux_to_I0_np",
    "sersic_Ie_to_flux_np",
    "sersic_flux_to_Ie_np",
    "sersic_I0_to_flux_torch",
    "sersic_flux_to_I0_torch",
    "sersic_Ie_to_flux_torch",
    "sersic_flux_to_Ie_torch",
    "sersic_inv_np",
    "sersic_inv_torch",
    "moffat_I0_to_flux",
)


def sersic_n_to_b(
    n: Union[float, np.ndarray, ArrayLike],
) -> Union[float, np.ndarray, ArrayLike]:
    """Compute the `b(n)` for a sersic model. This factor ensures that
    the $R_e$ and $I_e$ parameters do in fact correspond
    to the half light values and not some other scale
    radius/intensity.

    """

    return (
        2 * n
        - 1 / 3
        + 4 / (405 * n)
        + 46 / (25515 * n**2)
        + 131 / (1148175 * n**3)
        - 2194697 / (30690717750 * n**4)
    )


def sersic_I0_to_flux_np(I0: np.ndarray, n: np.ndarray, R: np.ndarray, q: np.ndarray) -> np.ndarray:
    """Compute the total flux integrated to infinity for a 2D elliptical
    sersic given the $I_0,n,R_s,q$ parameters which uniquely
    define the profile ($I_0$ is the central intensity in
    flux/arcsec^2). Note that $R_s$ is not the effective radius,
    but in fact the scale radius in the more straightforward sersic
    representation:

    $$I(R) = I_0e^{-(R/R_s)^{1/n}}$$

    **Args:**
    -  `I0`: central intensity (flux/arcsec^2)
    -  `n`: sersic index
    -  `R`: Scale radius
    -  `q`: axis ratio (b/a)

    """
    return 2 * np.pi * I0 * q * n * R**2 * gamma(2 * n)


def sersic_flux_to_I0_np(
    flux: np.ndarray, n: np.ndarray, R: np.ndarray, q: np.ndarray
) -> np.ndarray:
    """Compute the central intensity (flux/arcsec^2) for a 2D elliptical
    sersic given the $F,n,R_s,q$ parameters which uniquely
    define the profile ($F$ is the total flux integrated to
    infinity). Note that $R_s$ is not the effective radius, but
    in fact the scale radius in the more straightforward sersic
    representation:

    $$I(R) = I_0e^{-(R/R_s)^{1/n}}$$

    **Args:**
    -  `flux`: total flux integrated to infinity (flux)
    -  `n`: sersic index
    -  `R`: Scale radius
    -  `q`: axis ratio (b/a)

    """
    return flux / (2 * np.pi * q * n * R**2 * gamma(2 * n))


def sersic_Ie_to_flux_np(Ie: np.ndarray, n: np.ndarray, R: np.ndarray, q: np.ndarray) -> np.ndarray:
    """Compute the total flux integrated to infinity for a 2D elliptical
    sersic given the $I_e,n,R_e,q$ parameters which uniquely
    define the profile ($I_e$ is the intensity at $R_e$ in
    flux/arcsec^2). Note that $R_e$ is the effective radius in
    the sersic representation:

    $$I(R) = I_ee^{-b_n[(R/R_e)^{1/n}-1]}$$

    **Args:**
    -  `Ie`: intensity at the effective radius (flux/arcsec^2)
    -  `n`: sersic index
    -  `R`: Scale radius
    -  `q`: axis ratio (b/a)
    """
    bn = sersic_n_to_b(n)
    return 2 * np.pi * Ie * R**2 * q * n * (np.exp(bn) * bn ** (-2 * n)) * gamma(2 * n)


def sersic_flux_to_Ie_np(
    flux: np.ndarray, n: np.ndarray, R: np.ndarray, q: np.ndarray
) -> np.ndarray:
    """Compute the intensity at $R_e$ (flux/arcsec^2) for a 2D
    elliptical sersic given the $F,n,R_e,q$ parameters which
    uniquely define the profile ($F$ is the total flux
    integrated to infinity). Note that $R_e$ is the effective
    radius in the sersic representation:

    $$I(R) = I_ee^{-b_n[(R/R_e)^{1/n}-1]}$$

    **Args:**
    -  `flux`: flux integrated to infinity (flux)
    -  `n`: sersic index
    -  `R`: Scale radius
    -  `q`: axis ratio (b/a)

    """
    bn = sersic_n_to_b(n)
    return flux / (2 * np.pi * R**2 * q * n * (np.exp(bn) * bn ** (-2 * n)) * gamma(2 * n))


def sersic_inv_np(I: np.ndarray, n: np.ndarray, Re: np.ndarray, Ie: np.ndarray) -> np.ndarray:
    """Invert the sersic profile. Compute the radius corresponding to a
    given intensity for a pure sersic profile.

    """
    bn = sersic_n_to_b(n)
    return Re * ((1 - (1 / bn) * np.log(I / Ie)) ** (n))


def sersic_I0_to_flux_torch(I0: ArrayLike, n: ArrayLike, R: ArrayLike, q: ArrayLike) -> ArrayLike:
    """Compute the total flux integrated to infinity for a 2D elliptical
    sersic given the $I_0,n,R_s,q$ parameters which uniquely
    define the profile ($I_0$ is the central intensity in
    flux/arcsec^2). Note that $R_s$ is not the effective radius,
    but in fact the scale radius in the more straightforward sersic
    representation:

    $$I(R) = I_0e^{-(R/R_s)^{1/n}}$$

    **Args:**
    -  `I0`: central intensity (flux/arcsec^2)
    -  `n`: sersic index
    -  `R`: Scale radius
    -  `q`: axis ratio (b/a)


    """
    return 2 * np.pi * I0 * q * n * R**2 * backend.exp(backend.gammaln(2 * n))


def sersic_flux_to_I0_torch(flux: ArrayLike, n: ArrayLike, R: ArrayLike, q: ArrayLike) -> ArrayLike:
    """Compute the central intensity (flux/arcsec^2) for a 2D elliptical
    sersic given the $F,n,R_s,q$ parameters which uniquely
    define the profile ($F$ is the total flux integrated to
    infinity). Note that $R_s$ is not the effective radius, but
    in fact the scale radius in the more straightforward sersic
    representation:

    $$I(R) = I_0e^{-(R/R_s)^{1/n}}$$

    **Args:**
    -  `flux`: total flux integrated to infinity (flux)
    -  `n`: sersic index
    -  `R`: Scale radius
    -  `q`: axis ratio (b/a)

    """
    return flux / (2 * np.pi * q * n * R**2 * backend.exp(backend.gammaln(2 * n)))


def sersic_Ie_to_flux_torch(Ie: ArrayLike, n: ArrayLike, R: ArrayLike, q: ArrayLike) -> ArrayLike:
    """Compute the total flux integrated to infinity for a 2D elliptical
    sersic given the $I_e,n,R_e,q$ parameters which uniquely
    define the profile ($I_e$ is the intensity at $R_e$ in
    flux/arcsec^2). Note that $R_e$ is the effective radius in
    the sersic representation:

    $$I(R) = I_ee^{-b_n[(R/R_e)^{1/n}-1]}$$

    **Args:**
    -  `Ie`: intensity at the effective radius (flux/arcsec^2)
    -  `n`: sersic index
    -  `R`: Scale radius
    -  `q`: axis ratio (b/a)


    """
    bn = sersic_n_to_b(n)
    return (
        2
        * np.pi
        * Ie
        * R**2
        * q
        * n
        * (backend.exp(bn) * bn ** (-2 * n))
        * backend.exp(backend.gammaln(2 * n))
    )


def sersic_flux_to_Ie_torch(flux: ArrayLike, n: ArrayLike, R: ArrayLike, q: ArrayLike) -> ArrayLike:
    """Compute the intensity at $R_e$ (flux/arcsec^2) for a 2D
    elliptical sersic given the $F,n,R_e,q$ parameters which
    uniquely define the profile ($F$ is the total flux
    integrated to infinity). Note that $R_e$ is the effective
    radius in the sersic representation:

    $$I(R) = I_ee^{-b_n[(R/R_e)^{1/n}-1]}$$

    **Args:**
    -  `flux`: flux integrated to infinity (flux)
    -  `n`: sersic index
    -  `R`: Scale radius
    -  `q`: axis ratio (b/a)


    """
    bn = sersic_n_to_b(n)
    return flux / (
        2
        * np.pi
        * R**2
        * q
        * n
        * (backend.exp(bn) * bn ** (-2 * n))
        * backend.exp(backend.gammaln(2 * n))
    )


def sersic_inv_torch(I: ArrayLike, n: ArrayLike, Re: ArrayLike, Ie: ArrayLike) -> ArrayLike:
    """Invert the sersic profile. Compute the radius corresponding to a
    given intensity for a pure sersic profile.

    """
    bn = sersic_n_to_b(n)
    return Re * ((1 - (1 / bn) * backend.log(I / Ie)) ** (n))


def moffat_I0_to_flux(I0: float, n: float, rd: float, q: float) -> float:
    """
    Compute the total flux integrated to infinity for a moffat profile.

    **Args:**
    -  `I0`: central intensity (flux/arcsec^2)
    -  `n`: moffat curvature parameter (unitless)
    -  `rd`: scale radius
    -  `q`: axis ratio
    """
    return I0 * np.pi * rd**2 * q / (n - 1)



from typing import Sequence, Literal

from scipy.optimize import minimize
import numpy as np

from .base import BaseOptimizer
from .. import config
from ..backend_obj import backend

__all__ = ("ScipyFit",)


class ScipyFit(BaseOptimizer):
    """Scipy-based optimizer for fitting models to data using various
    optimization methods.

    The optimizer uses the `scipy.optimize.minimize` function to perform the
    fitting. The Scipy package is widely used and well tested for optimization
    tasks. It supports a variety of methods, however only a subset allow users to
    define boundaries for the parameters. This wrapper is only for those methods.

    **Args:**
    -  `model`: The model to fit, which should be an instance of `Model`.
    -  `initial_state`: Initial guess for the model parameters as a 1D tensor.
    -  `method`: The optimization method to use. Default is "Nelder-Mead", but can be set to any of: "Nelder-Mead", "L-BFGS-B", "TNC", "SLSQP", "Powell", or "trust-constr".
    -  `ndf`: Optional number of degrees of freedom for the fit. If not provided, it is calculated as the number of data points minus the number of parameters.
    """

    def __init__(
        self,
        model,
        initial_state: Sequence = None,
        method: Literal[
            "Nelder-Mead", "L-BFGS-B", "TNC", "SLSQP", "Powell", "trust-constr"
        ] = "Nelder-Mead",
        likelihood: Literal["gaussian", "poisson"] = "gaussian",
        ndf=None,
        **kwargs,
    ):

        super().__init__(model, initial_state, **kwargs)
        self.method = method
        self.likelihood = likelihood

        # Degrees of freedom
        if ndf is None:
            sub_target = self.model.target[self.model.window]
            ndf = (
                np.prod(sub_target.flatten("data").shape)
                - backend.sum(sub_target.flatten("mask")).item()
            )
            self.ndf = max(1.0, ndf - len(self.current_state))
        else:
            self.ndf = ndf

    def numpy_bounds(self):
        """Convert the model's parameter bounds to a format suitable for scipy.optimize."""
        bounds = []
        for param in self.model.dynamic_params:
            if param.shape == ():
                bound = [None, None]
                if param.valid[0] is not None:
                    bound[0] = backend.to_numpy(param.valid[0])
                if param.valid[1] is not None:
                    bound[1] = backend.to_numpy(param.valid[1])
                bounds.append(tuple(bound))
            else:
                for i in range(np.prod(param.value.shape)):
                    bound = [None, None]
                    if param.valid[0] is not None:
                        bound[0] = backend.to_numpy(param.valid[0].flatten()[i])
                    if param.valid[1] is not None:
                        bound[1] = backend.to_numpy(param.valid[1].flatten()[i])
                    bounds.append(tuple(bound))
        return bounds

    def density(self, state: Sequence) -> float:
        if self.likelihood == "gaussian":
            return -self.model.gaussian_log_likelihood(
                backend.as_array(state, dtype=config.DTYPE, device=config.DEVICE)
            ).item()
        elif self.likelihood == "poisson":
            return -self.model.poisson_log_likelihood(
                backend.as_array(state, dtype=config.DTYPE, device=config.DEVICE)
            ).item()
        else:
            raise ValueError(f"Unknown likelihood type: {self.likelihood}")

    def fit(self):

        res = minimize(
            lambda x: self.density(x),
            self.current_state,
            method=self.method,
            bounds=self.numpy_bounds(),
            options={
                "maxiter": self.max_iter,
            },
        )
        self.scipy_res = res
        self.message = self.message + f"success: {res.success}, message: {res.message}"
        self.current_state = backend.as_array(res.x, dtype=config.DTYPE, device=config.DEVICE)
        if self.verbose > 0:
            config.logger.info(
                f"Final 2NLL/DoF: {2*self.density(res.x)/self.ndf:.6g}. Converged: {self.message}"
            )
        self.model.set_values(self.current_state)

        return self



import torch
import numpy as np

from ..backend_obj import backend, ArrayLike

__all__ = ("default_prof", "interp2d")


def default_prof(
    shape: tuple[int, int], pixelscale: float, min_pixels: int = 2, scale: float = 0.2
) -> np.ndarray:
    prof = [0, min_pixels * pixelscale]
    imagescale = max(shape)  # np.sqrt(np.sum(np.array(shape) ** 2))
    while prof[-1] < (imagescale * pixelscale / 2):
        prof.append(prof[-1] + max(min_pixels * pixelscale, prof[-1] * scale))
    return np.array(prof)


def interp2d(
    im: ArrayLike,
    i: ArrayLike,
    j: ArrayLike,
    padding_mode: str = "zeros",
) -> ArrayLike:
    """
    Interpolates a 2D image at specified coordinates.
    Similar to `torch.nn.functional.grid_sample` with `align_corners=False`.

    Args:
        im (Tensor): A 2D tensor representing the image.
        i (Tensor): A tensor of i coordinates (in pixel space) at which to interpolate.
        j (Tensor): A tensor of j coordinates (in pixel space) at which to interpolate.

    Returns:
        Tensor: Tensor with the same shape as `i` and `j` containing the interpolated values.
    """

    # Convert coordinates to pixel indices
    h, w = im.shape

    # reshape for indexing purposes
    start_shape = i.shape
    i = i.flatten()
    j = j.flatten()

    # valid
    valid = (i >= -0.5) & (i <= (h - 0.5)) & (j >= -0.5) & (j <= (w - 0.5))

    i0 = backend.long(backend.floor(i))
    j0 = backend.long(backend.floor(j))
    i0 = backend.clamp(i0, 0, h - 2)
    i1 = i0 + 1
    j0 = backend.clamp(j0, 0, w - 2)
    j1 = j0 + 1

    fa = im[i0, j0]
    fb = im[i0, j1]
    fc = im[i1, j0]
    fd = im[i1, j1]

    wa = (i1 - i) * (j1 - j)
    wb = (i1 - i) * (j - j0)
    wc = (i - i0) * (j1 - j)
    wd = (i - i0) * (j - j0)

    result = fa * wa + fb * wb + fc * wc + fd * wd

    if padding_mode == "zeros":
        return (result * valid).reshape(start_shape)
    elif padding_mode == "border":
        return result.reshape(start_shape)
    raise ValueError(f"Unsupported padding mode: {padding_mode}")



import numpy as np
from tqdm import tqdm


def mala(
    initial_state,  # (num_chains, D)
    log_prob,  # (num_chains, D) -> (num_chains,)
    log_prob_grad,  # (num_chains, D) -> (num_chains, D)
    num_samples,
    epsilon,
    mass_matrix,  # covariance
    progress=True,
    desc="MALA",
):
    x = np.array(initial_state, copy=True)
    C, D = x.shape

    # mass, inv_mass, L
    mass = np.array(mass_matrix, copy=False)  # (D, D)
    inv_mass = np.linalg.inv(mass)  # (D, D)
    L = np.linalg.cholesky(mass)  # (D, D)

    samples = np.zeros((num_samples, C, D), dtype=x.dtype)  # (N, C, D)
    acceptance_rate = np.zeros([0])  # (0,)
    logp = np.zeros((num_samples, C), dtype=x.dtype)  # (N, C)

    # Cache current state
    logp_cur = log_prob(x)  # (C,)
    grad_cur = log_prob_grad(x)  # (C, D)

    # Random number generator
    rng = np.random.default_rng(np.random.randint(1e9))

    it = range(num_samples)
    if progress:
        it = tqdm(it, desc=desc, position=0, leave=True)

    for t in it:
        # proposal using current grad
        mu_x = 0.5 * (epsilon**2) * (grad_cur @ mass)  # (C, D)
        noise = rng.standard_normal((C, D)) @ L.T  # (C, D)
        x_prop = x + mu_x + epsilon * noise  # (C, D)

        # Evaluate proposal
        logp_prop = log_prob(x_prop)  # (C,)
        grad_prop = log_prob_grad(x_prop)  # (C, D)

        mu_xprop = 0.5 * (epsilon**2) * (grad_prop @ mass)  # (C, D)

        # q(x|x') \propto \exp(-0.5|x - x' - mu(x')|^2 / \epsilon^2)
        d1 = x - x_prop - mu_xprop  # for q(x | x')
        d2 = x_prop - x - mu_x  # for q(x'| x)

        logq1 = -0.5 * np.einsum("bi,ij,bj->b", d1, inv_mass, d1) / epsilon**2  # (C,)
        logq2 = -0.5 * np.einsum("bi,ij,bj->b", d2, inv_mass, d2) / epsilon**2  # (C,)

        log_alpha = (logp_prop - logp_cur) + (logq1 - logq2)  # (C,)

        accept = np.log(rng.random(C)) < log_alpha  # (C,)
        acceptance_rate = np.concatenate([acceptance_rate, accept])

        # Update all three pieces in-place where accepted
        x[accept] = x_prop[accept]  # (C, D)
        logp_cur[accept] = logp_prop[accept]  # (C,)
        grad_cur[accept] = grad_prop[accept]  # (C, D)

        samples[t] = x.copy()
        logp[t] = logp_cur.copy()

        if progress:
            it.set_postfix(acc_rate=f"{acceptance_rate.mean():0.2f}")

    return samples, logp



import torch
import numpy as np

from ..utils.decorators import ignore_numpy_warnings, combine_docstrings
from .psf_model_object import PSFModel
from .mixins import RadialMixin
from ..param import forward
from ..backend_obj import backend, ArrayLike

__all__ = ("AiryPSF",)


@combine_docstrings
class AiryPSF(RadialMixin, PSFModel):
    """The Airy disk is an analytic description of the diffraction pattern
    for a circular aperture.

    The diffraction pattern is described exactly by the configuration
    of the lens system under the assumption that all elements are
    perfect. This expression goes as:

    $$I(\\theta) = I_0\\left[\\frac{2J_1(x)}{x}\\right]^2$$
    $$x = ka\\sin(\\theta) = \\frac{2\\pi a r}{\\lambda R}$$

    where $I(\\theta)$ is the intensity as a function of the
    angular position within the diffraction system along its main
    axis, $I_0$ is the central intensity of the airy disk,
    $J_1$ is the Bessel function of the first kind of order one,
    $k = \\frac{2\\pi}{\\lambda}$ is the wavenumber of the
    light, $a$ is the aperture radius, $r$ is the radial
    position from the center of the pattern, $R$ is the distance
    from the circular aperture to the observation plane.

    In the `Airy_PSF` class we combine the parameters
    $a,R,\\lambda$ into a single ratio to be optimized (or fixed
    by the optical configuration).

    **Parameters:**
    -    `I0`: The central intensity of the airy disk in flux/arcsec^2.
    -    `aRL`: The ratio of the aperture radius to the
        product of the wavelength and the distance from the aperture to the
        observation plane, $\\frac{a}{R \\lambda}$.

    """

    _model_type = "airy"
    _parameter_specs = {
        "I0": {"units": "flux/arcsec^2", "value": 1.0, "shape": (), "dynamic": False},
        "aRL": {"units": "a/(R lambda)", "shape": (), "dynamic": True},
    }
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if self.I0.initialized and self.aRL.initialized:
            return
        icenter = self.target.plane_to_pixel(*self.center.value)

        if not self.I0.initialized:
            mid_chunk = self.target._data[
                int(icenter[0]) - 2 : int(icenter[0]) + 2,
                int(icenter[1]) - 2 : int(icenter[1]) + 2,
            ]
            self.I0.value = backend.mean(mid_chunk) / self.target.pixel_area
        if not self.aRL.initialized:
            self.aRL.value = (5.0 / 8.0) * 2 * self.target.pixelscale

    @forward
    def radial_model(self, R: ArrayLike, I0: ArrayLike, aRL: ArrayLike) -> ArrayLike:
        x = 2 * np.pi * aRL * R
        return I0 * (2 * backend.bessel_j1(x) / x) ** 2



import os
import importlib
from typing import Annotated

from torch import Tensor, dtype, device
import torch
import numpy as np
import caskade as ck

from . import config

ArrayLike = Annotated[
    Tensor,
    "One of: torch.Tensor or jax.numpy.ndarray depending on the chosen backend.",
]
dtypeLike = Annotated[
    dtype,
    "One of: torch.dtype or jax.numpy.dtype depending on the chosen backend.",
]
deviceLike = Annotated[
    device,
    "One of: torch.device or jax.DeviceArray depending on the chosen backend.",
]


class Backend:
    def __init__(self, backend=None):
        self.backend = backend

    @property
    def backend(self):
        return self._backend

    @backend.setter
    def backend(self, backend):
        if backend is None:
            backend = os.getenv("CASKADE_BACKEND", "torch")
        ck.backend.backend = backend
        self._load_backend(backend)
        self._backend = backend

    def _load_backend(self, backend):
        if backend == "torch":
            self.module = importlib.import_module("torch")
            self.setup_torch()
        elif backend == "jax":
            self.module = importlib.import_module("jax.numpy")
            self.setup_jax()
        else:
            raise ValueError(f"Unsupported backend: {backend}")

    def setup_torch(self):
        config.DTYPE = torch.float64
        config.DEVICE = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.make_array = self._make_array_torch
        self._array_type = self._array_type_torch
        self.concatenate = self._concatenate_torch
        self.copy = self._copy_torch
        self.tolist = self._tolist_torch
        self.view = self._view_torch
        self.as_array = self._as_array_torch
        self.to = self._to_torch
        self.to_numpy = self._to_numpy_torch
        self.gammaln = self._gammaln_torch
        self.logit = self._logit_torch
        self.sigmoid = self._sigmoid_torch
        self.repeat = self._repeat_torch
        self.stack = self._stack_torch
        self.transpose = self._transpose_torch
        self.upsample2d = self._upsample2d_torch
        self.pad = self._pad_torch
        self.LinAlgErr = self.module._C._LinAlgError
        self.roll = self._roll_torch
        self.clamp = self._clamp_torch
        self.flatten = self._flatten_torch
        self.conv2d = self._conv2d_torch
        self.mean = self._mean_torch
        self.sum = self._sum_torch
        self.max = self._max_torch
        self.topk = self._topk_torch
        self.bessel_j1 = self._bessel_j1_torch
        self.bessel_k1 = self._bessel_k1_torch
        self.lgamma = self._lgamma_torch
        self.hessian = self._hessian_torch
        self.jacobian = self._jacobian_torch
        self.jacfwd = self._jacfwd_torch
        self.grad = self._grad_torch
        self.vmap = self._vmap_torch
        self.long = self._long_torch
        self.detach = lambda x: x.detach()
        self.fill_at_indices = self._fill_at_indices_torch
        self.add_at_indices = self._add_at_indices_torch
        self.and_at_indices = self._and_at_indices_torch

    def setup_jax(self):
        self.jax = importlib.import_module("jax")
        self.jax.config.update("jax_enable_x64", True)
        config.DTYPE = None
        config.DEVICE = None
        self.make_array = self._make_array_jax
        self._array_type = self._array_type_jax
        self.concatenate = self._concatenate_jax
        self.copy = self._copy_jax
        self.tolist = self._tolist_jax
        self.view = self._view_jax
        self.as_array = self._as_array_jax
        self.to = self._to_jax
        self.to_numpy = self._to_numpy_jax
        self.gammaln = self._gammaln_jax
        self.logit = self._logit_jax
        self.sigmoid = self._sigmoid_jax
        self.repeat = self._repeat_jax
        self.stack = self._stack_jax
        self.transpose = self._transpose_jax
        self.upsample2d = self._upsample2d_jax
        self.pad = self._pad_jax
        self.LinAlgErr = Exception
        self.roll = self._roll_jax
        self.clamp = self._clamp_jax
        self.flatten = self._flatten_jax
        self.conv2d = self._conv2d_jax
        self.mean = self._mean_jax
        self.sum = self._sum_jax
        self.max = self._max_jax
        self.topk = self._topk_jax
        self.bessel_j1 = self._bessel_j1_jax
        self.bessel_k1 = self._bessel_k1_jax
        self.lgamma = self._lgamma_jax
        self.hessian = self._hessian_jax
        self.jacobian = self._jacobian_jax
        self.jacfwd = self._jacfwd_jax
        self.grad = self._grad_jax
        self.vmap = self._vmap_jax
        self.long = self._long_jax
        self.detach = lambda x: x
        self.fill_at_indices = self._fill_at_indices_jax
        self.add_at_indices = self._add_at_indices_jax
        self.and_at_indices = self._and_at_indices_jax

    @property
    def array_type(self):
        return self._array_type()

    def _make_array_torch(self, array, dtype=None, device=None):
        return self.module.tensor(array, dtype=dtype, device=device)

    def _make_array_jax(self, array, dtype=None, **kwargs):
        return self.module.array(array, dtype=dtype)

    def _array_type_torch(self):
        return self.module.Tensor

    def _array_type_jax(self):
        return self.module.ndarray

    def _concatenate_torch(self, arrays, dim=0):
        return self.module.cat(arrays, dim=dim)

    def _concatenate_jax(self, arrays, dim=0):
        return self.module.concatenate(arrays, axis=dim)

    def _copy_torch(self, array):
        return array.detach().clone()

    def _copy_jax(self, array):
        return self.module.copy(array)

    def _tolist_torch(self, array):
        return array.detach().cpu().tolist()

    def _tolist_jax(self, array):
        return array.block_until_ready().tolist()

    def _view_torch(self, array, shape):
        return array.reshape(shape)

    def _view_jax(self, array, shape):
        return array.reshape(shape)

    def _as_array_torch(self, array, dtype=None, device=None):
        return self.module.as_tensor(array, dtype=dtype, device=device)

    def _as_array_jax(self, array, dtype=None, **kwargs):
        return self.module.asarray(array, dtype=dtype)

    def _to_torch(self, array, dtype=None, device=None):
        return array.to(dtype=dtype, device=device)

    def _to_jax(self, array, dtype=None, device=None):
        return self.jax.device_put(array.astype(dtype), device=device)

    def _to_numpy_torch(self, array):
        return array.detach().cpu().numpy()

    def _to_numpy_jax(self, array):
        return np.array(array.block_until_ready())

    def _repeat_torch(self, a, repeats, axis=None):
        return self.module.repeat_interleave(a, repeats, dim=axis)

    def _repeat_jax(self, a, repeats, axis=None):
        return self.module.repeat(a, repeats, axis=axis)

    def _stack_torch(self, arrays, dim=0):
        return self.module.stack(arrays, dim=dim)

    def _stack_jax(self, arrays, dim=0):
        return self.module.stack(arrays, axis=dim)

    def _transpose_torch(self, array, *args):
        return self.module.transpose(array, *args)

    def _transpose_jax(self, array, *args):
        permutation = np.arange(array.ndim)
        permutation[np.sort(args)] = args
        return self.module.transpose(array, permutation)

    def _gammaln_torch(self, array):
        return self.module.special.gammaln(array)

    def _gammaln_jax(self, array):
        return self.jax.scipy.special.gammaln(array)

    def _sigmoid_torch(self, array):
        return self.module.sigmoid(array)

    def _sigmoid_jax(self, array):
        return self.jax.nn.sigmoid(array)

    def _logit_torch(self, array):
        return self.module.logit(array)

    def _logit_jax(self, array):
        return self.jax.scipy.special.logit(array)

    def _upsample2d_torch(self, array, scale_factor, method):
        U = self.module.nn.Upsample(scale_factor=scale_factor, mode=method)
        array = U(array) / scale_factor**2
        return array

    def _upsample2d_jax(self, array, scale_factor, method):
        if method == "nearest":
            method = "bilinear"  # no nearest neighbor interpolation in jax
        new_shape = list(array.shape)
        new_shape[-2] = array.shape[-2] * scale_factor
        new_shape[-1] = array.shape[-1] * scale_factor
        return self.jax.image.resize(array, new_shape, method=method)

    def _pad_torch(self, array, padding, mode):
        return self.module.nn.functional.pad(array, padding[-4:], mode=mode)

    def _pad_jax(self, array, padding, mode):
        if mode == "replicate":
            mode = "edge"
        padding = np.array(padding).reshape(-1, 2)
        return self.module.pad(array, padding, mode=mode)

    def _roll_torch(self, array, shifts, dims):
        return self.module.roll(array, shifts, dims=dims)

    def _roll_jax(self, array, shifts, dims):
        return self.module.roll(array, shifts, axis=dims)

    def _clamp_torch(self, array, min, max):
        return self.module.clamp(array, min, max)

    def _clamp_jax(self, array, min, max):
        return self.module.clip(array, min, max)

    def _long_torch(self, array):
        return array.long()

    def _long_jax(self, array):
        return self.module.astype(array, self.module.int64)

    def _conv2d_torch(self, input, kernel, padding, stride=1):
        return self.module.nn.functional.conv2d(
            input,
            kernel,
            padding=padding,
            stride=stride,
        )

    def _conv2d_jax(self, input, kernel, padding, stride=1):
        return self.jax.lax.conv_general_dilated(
            input, kernel, window_strides=(stride, stride), padding=padding
        )

    def _mean_torch(self, array, dim=None):
        return self.module.mean(array, dim=dim)

    def _mean_jax(self, array, dim=None):
        return self.module.mean(array, axis=dim)

    def _sum_torch(self, array, dim=None):
        return self.module.sum(array, dim=dim)

    def _sum_jax(self, array, dim=None):
        return self.module.sum(array, axis=dim)

    def _max_torch(self, array, dim=None):
        return array.amax(dim=dim)

    def _max_jax(self, array, dim=None):
        return self.module.max(array, axis=dim)

    def _topk_torch(self, array, k):
        return self.module.topk(array, k=k)

    def _topk_jax(self, array, k):
        return self.jax.lax.top_k(array, k=k)

    def _bessel_j1_torch(self, array):
        return self.module.special.bessel_j1(array)

    def _bessel_j1_jax(self, array):
        return self.jax.scipy.special.bessel_jn(array, v=1)[-1]

    def _bessel_k1_torch(self, array):
        return self.module.special.modified_bessel_k1(array)

    def _bessel_k1_jax(self, array):
        return self.jax.scipy.special.kn(1, array)

    def _lgamma_torch(self, array):
        return self.module.lgamma(array)

    def _lgamma_jax(self, array):
        return self.jax.lax.lgamma(array)

    def _grad_torch(self, func):
        return self.module.func.grad(func)

    def _grad_jax(self, func):
        return self.jax.grad(func)

    def _jacobian_torch(self, func, x, strategy="forward-mode", vectorize=True, create_graph=False):
        return self.module.autograd.functional.jacobian(
            func, x, strategy=strategy, vectorize=vectorize, create_graph=create_graph
        )

    def _jacobian_jax(self, func, x, strategy="forward-mode", vectorize=True, create_graph=False):
        if "forward" in strategy:
            # n = x.size
            # eye = self.module.eye(n)
            # Jt = self.jax.vmap(lambda s: self.jax.jvp(func, (x,), (s,))[1])(eye)
            # return self.module.moveaxis(Jt, 0, -1)
            return self.jax.jacfwd(func)(x)
        return self.jax.jacrev(func)(x)

    def _jacfwd_torch(self, func, argnums=0):
        return self.module.func.jacfwd(func, argnums=argnums)

    def _jacfwd_jax(self, func, argnums=0):
        return self.jax.jacfwd(func, argnums=argnums)

    def _hessian_torch(self, func):
        return self.module.func.hessian(func)

    def _hessian_jax(self, func):
        return self.jax.hessian(func)

    def _vmap_torch(self, *args, **kwargs):
        return self.module.vmap(*args, **kwargs)

    def _vmap_jax(self, *args, **kwargs):
        return self.jax.vmap(*args, **kwargs)

    def _fill_at_indices_torch(self, array, indices, values):
        array[indices] = values
        return array

    def _fill_at_indices_jax(self, array, indices, values):
        array = array.at[indices].set(values)
        return array

    def _add_at_indices_torch(self, array, indices, values):
        array[indices] += values
        return array

    def _add_at_indices_jax(self, array, indices, values):
        array = array.at[indices].add(values)
        return array

    def _and_at_indices_torch(self, array, indices, values):
        array[indices] &= values
        return array

    def _and_at_indices_jax(self, array, indices, values):
        array = array.at[indices].set(array[indices] & values)
        return array

    def _flatten_torch(self, array, start_dim=0, end_dim=-1):
        return array.flatten(start_dim, end_dim)

    def _flatten_jax(self, array, start_dim=0, end_dim=-1):
        shape = tuple(array.shape)
        end_dim = (end_dim % len(shape)) + 1
        new_shape = shape[:start_dim] + (-1,) + shape[end_dim:]
        return self.module.reshape(array, new_shape)

    def arange(self, *args, dtype=None, device=None):
        return self.module.arange(*args, dtype=dtype, device=device)

    def linspace(self, start, end, steps, dtype=None, device=None):
        return self.module.linspace(start, end, steps, dtype=dtype, device=device)

    def meshgrid(self, *arrays, indexing="ij"):
        return self.module.meshgrid(*arrays, indexing=indexing)

    def searchsorted(self, array, value):
        return self.module.searchsorted(array, value)

    def any(self, array):
        return self.module.any(array)

    def all(self, array):
        return self.module.all(array)

    def log(self, array):
        return self.module.log(array)

    def log10(self, array):
        return self.module.log10(array)

    def exp(self, array):
        return self.module.exp(array)

    def sin(self, array):
        return self.module.sin(array)

    def cos(self, array):
        return self.module.cos(array)

    def cosh(self, array):
        return self.module.cosh(array)

    def sqrt(self, array):
        return self.module.sqrt(array)

    def abs(self, array):
        return self.module.abs(array)

    def floor(self, array):
        return self.module.floor(array)

    def tanh(self, array):
        return self.module.tanh(array)

    def arctan(self, array):
        return self.module.arctan(array)

    def arctan2(self, y, x):
        return self.module.arctan2(y, x)

    def arcsin(self, array):
        return self.module.arcsin(array)

    def round(self, array):
        return self.module.round(array)

    def zeros(self, shape, dtype=None, device=None):
        return self.module.zeros(shape, dtype=dtype, device=device)

    def zeros_like(self, array, dtype=None):
        return self.module.zeros_like(array, dtype=dtype)

    def ones(self, shape, dtype=None, device=None):
        return self.module.ones(shape, dtype=dtype, device=device)

    def ones_like(self, array, dtype=None):
        return self.module.ones_like(array, dtype=dtype)

    def empty(self, shape, dtype=None, device=None):
        return self.module.empty(shape, dtype=dtype, device=device)

    def eye(self, n, dtype=None, device=None):
        return self.module.eye(n, dtype=dtype, device=device)

    def diag(self, array):
        return self.module.diag(array)

    def outer(self, a, b):
        return self.module.outer(a, b)

    def minimum(self, a, b):
        return self.module.minimum(a, b)

    def maximum(self, a, b):
        return self.module.maximum(a, b)

    def isnan(self, array):
        return self.module.isnan(array)

    def isfinite(self, array):
        return self.module.isfinite(array)

    def where(self, condition, x, y):
        return self.module.where(condition, x, y)

    def allclose(self, a, b, rtol=1e-5, atol=1e-8):
        return self.module.allclose(a, b, rtol=rtol, atol=atol)

    @property
    def linalg(self):
        return self.module.linalg

    @property
    def fft(self):
        return self.module.fft

    @property
    def inf(self):
        return self.module.inf

    @property
    def bool(self):
        return self.module.bool

    @property
    def int32(self):
        return self.module.int32

    @property
    def float32(self):
        return self.module.float32

    @property
    def float64(self):
        return self.module.float64


backend = Backend()
